
#if defined(__WINDOWS__) || defined(__NT__) || defined(_MSC_VER) || defined(_WIN32)
#       define __WIN32__
#endif


#include <stdio.h>

#ifdef  __WIN32__
#       define WIN32_LEAN_AND_MEAN
//#		include <conio.h>
#       include <windows.h>
#       include <winsock.h>
#else
#       include <sys/types.h>
#       include <sys/socket.h>
#       include <netinet/in.h>
#       include <arpa/inet.h>
#       include <errno.h>
#       include <unistd.h>
#       include <netdb.h>
#       include <sys/ioctl.h>
#       include <sys/time.h>
#endif

#ifndef __WIN32__
typedef int SOCKET;
#define SOCKET_ERROR -1
#define INVALID_SOCKET -1
#define closesocket close
#define ioctlsocket ioctl
#define Sleep(x)        usleep (x * 1000)
#endif

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <ctype.h>
#include <math.h>

#include "doomtype.h"
#include "doomstat.h"
#include "m_argv.h"

#include "cl_main.h"
#include "cl_huffman.h"
#include "m_random.h"


int         net_socket;
int         localport;
netadr_t    net_from;   // address of who sent the packet
netadr_t	net_local_adr;

// Our standard buffer where packets go to to be read  =)
sizebuf_t   msg_read;

#ifdef	WIN32
WSADATA		winsockdata;
#endif


// Client side only
void	STACK_ARGS Network_Error(char	*s,...);

//EXTERN_CVAR(Int, maxrate);

//
// ZD_UDPsocket
//
SOCKET ZD_UDPsocket (void)
{
        SOCKET s;

        // allocate a socket
        s = socket (PF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if (s == INVALID_SOCKET)
		{
			//I_FatalError ("can't create socket");
			Network_Error("cant create socket!\n");
		}

        return s;
}

//
// ZD_BindToLocalPort
//
void ZD_BindToLocalPort (SOCKET s, u_short port)
{
        int v;
        struct sockaddr_in address;

        memset (&address, 0, sizeof(address));
        address.sin_family = AF_INET;
        address.sin_addr.s_addr = INADDR_ANY;
        address.sin_port = htons(port);

        v = bind (s, (sockaddr *)&address, sizeof(address));
        if (v == SOCKET_ERROR)
		{
           Network_Error("BindToPort: error\n");
		}
}		   


void ZD_CloseNetwork (void)
{
	closesocket (net_socket);
#ifdef __WIN32__
    WSACleanup ();
#endif
}

void SockadrToNetadr (struct sockaddr_in *s, netadr_t *a)
{
     *(int *)&a->ip = *(int *)&s->sin_addr;
     a->port = s->sin_port;
}

void NetadrToSockadr (netadr_t *a, struct sockaddr_in *s)
{
     memset (s, 0, sizeof(*s));
     s->sin_family = AF_INET;

     *(int *)&s->sin_addr = *(int *)&a->ip;
     s->sin_port = a->port;
}

char *NET_AdrToString(netadr_t *a)
{
     static  char    s[64];

     sprintf (s, "%i.%i.%i.%i:%i", a->ip[0], a->ip[1], a->ip[2], a->ip[3], ntohs(a->port));

     return s;
}

bool NET_StringToAdr (char *s, netadr_t *a)
{
     struct hostent  *h;
     struct sockaddr_in sadr;
     char    *colon;
     char    copy[128];


     memset (&sadr, 0, sizeof(sadr));
     sadr.sin_family = AF_INET;

     sadr.sin_port = 0;

     strcpy (copy, s);
     // strip off a trailing :port if present
     for (colon = copy ; *colon ; colon++)
          if (*colon == ':')
          {
             *colon = 0;
             sadr.sin_port = htons(atoi(colon+1));
          }

     if (copy[0] >= '0' && copy[0] <= '9')
     {
          *(int *)&sadr.sin_addr = inet_addr(copy);
     }
     else
     {
          if (! (h = gethostbyname(copy)) )
                return 0;
          *(int *)&sadr.sin_addr = *(int *)h->h_addr_list[0];
     }

     SockadrToNetadr (&sadr, a);

     return true;
}

bool NET_CompareAdr(netadr_t *a, netadr_t *b)
{
    return (!memcmp(a->ip,b->ip,4) && a->port == b->port);
}



static unsigned char huffbuff[65536];

int ZD_GetPackets (void)
{
	int 	ret;
	struct sockaddr_in	from;
	int		fromlen;

	if (net_demoplayback)
		return 0;

    fromlen = sizeof(from);
#ifdef	WIN32
    ret = recvfrom (net_socket, (char *)huffbuff, sizeof(huffbuff), 0, (struct sockaddr *)&from, &fromlen);
#else
	ret = recvfrom (net_socket, (char *)huffbuff, sizeof(huffbuff), 0, (struct sockaddr *)&from, (socklen_t *)&fromlen);
#endif

    if (ret == -1) 
    { 
#ifdef __WIN32__
        errno = WSAGetLastError();

        if (errno == WSAEWOULDBLOCK)
            return false;
		
		// connection reset by peer...dosent mean anything to the server
		if (errno == WSAECONNRESET)
			return false;

        if (errno == WSAEMSGSIZE) 
		{
             Printf ("Warning:  Oversize packet from %s\n",
                      NET_AdrToString (&net_from));
             return false;
        }

        printf ("ZD_GetPackets(%d) Error(%d): %s\n", ret, errno, strerror(errno));
		return false;
#else
        if (errno == EWOULDBLOCK)
            return false;
        if (errno == ECONNREFUSED)
            return false;
	    
        Printf ("ZD_GetPackets: %s\n", strerror(errno));
        return false;
#endif
    }

	// No packets or an error, dont process anything.
	if (ret <= 0)
		return 0;

	// Compressed
	zd_inbytes += msg_read.cursize;

	HuffDecode(huffbuff, (unsigned char *)msg_read.data,ret,&ret);

	msg_read.readcount = 0;
	msg_read.cursize = ret;
	msg_read.compsize = ret;
	
    SockadrToNetadr (&from, &net_from);

	return msg_read.cursize;
}



void ZD_LaunchPacket (sizebuf_t netbuf, netadr_t to)
{
	int ret;
	struct sockaddr_in	addr;
	//int		 r;
	int		outlen;

	// [NightFang]- make sure we have something to send and
	// this isnt a "simulated" netgame
	if (netbuf.cursize <= 0 || net_demoplayback)
		return;

	// Packet loss emulation going out
	//if (P_Random(pr_misc) < 170 && gamestate == GS_LEVEL)
	//	return;

	//zd_uncompressedoutbytes += netbuf.cursize;
	
    NetadrToSockadr (&to, &addr);

	// Not using any more gay-ass huffman encoding
	HuffEncode((unsigned char *)netbuf.data,huffbuff,netbuf.cursize,&outlen);

	ret = sendto (net_socket, (const char*)huffbuff, outlen, 0, (struct sockaddr *)&addr, sizeof(addr));
    
    if (ret == -1) 
    {
#ifdef __WIN32__
          int err = WSAGetLastError();

          // wouldblock is silent
          if (err == WSAEWOULDBLOCK)
              return;
#else	  
          if (errno == EWOULDBLOCK)
              return;
          if (errno == ECONNREFUSED)
              return;
          Printf ("NET_SendPacket: %s\n", strerror(errno));
#endif	  
    }

	zd_outbytes += outlen;
}



//
// NET_GetLocalAddress - Old stuff
//

void NET_GetLocalAddress (void)
{
	char	buff[512];
	struct sockaddr_in	address;
	int		namelen;

	gethostname(buff, 512);
	buff[512-1] = 0;

	NET_StringToAdr (buff, &net_local_adr);

	namelen = sizeof(address);
	if (getsockname (net_socket, (struct sockaddr *)&address, &namelen) == -1)
	{
		Network_Error("NET_Init: getsockname:", strerror(errno));
	}
	net_local_adr.port = address.sin_port;

	Printf("IP address %s\n", NET_AdrToString (&net_local_adr));
}

//
// InitNetCommon
//
void InitNetCommon(void)
{
	int		r;
   unsigned long _true = true;	

#ifdef __WIN32__
   WSADATA   wsad;
   r = WSAStartup( 0x0101, &wsad );

   if (r)
	{	Network_Error("Winsock initialization failed!!\n"); }
#endif

   net_socket = ZD_UDPsocket ();
   ZD_BindToLocalPort (net_socket, localport);   
   if (ioctlsocket (net_socket, FIONBIO, &_true) == -1)
       Printf ("ZD_UDPsocket: ioctl FIONBIO:", strerror(errno));

   SZ_Init(&localbuf, MAX_UDP_PACKET);
   SZ_Clear(&localbuf);

   SZ_Init(&msg_read, MAX_UDP_PACKET);
   SZ_Clear(&msg_read);

	//
	// determine my name & address
	//
	NET_GetLocalAddress ();

	Printf(PRINT_HIGH, "UDP Initialized\n");
	
	// Just make sure they are disabled by default
	net_demorecording = net_demoplayback = demo_localinfo = false;
}


void I_SetPort(netadr_t &addr, int port)
{
   addr.port = htons(port);
}



void ZD_StartNetwork(void)
{
	HuffInit();

	// Set the default port
	localport = CLIENT_PORT;

	// No? Okay, use a custom -port
	const char *val = Args.CheckValue ("-port");
	if (val)
	{
		localport = atoi(val);
		Printf("Using alternate port %d\n", localport);
	}

	InitNetCommon();
}


void SZ_Init (sizebuf_t *buf, int length)
{
	memset (buf, 0, sizeof(*buf));
	//buf->data = new byte[length];
	buf->maxsize = length;
}

void SZ_Clear (sizebuf_t *buf)
{
	buf->cursize = 0;
	//buf->overflowed = false;
}

byte *SZ_GetSpace (sizebuf_t *buf, int length)
{
	byte	*data;
	
	if (buf->cursize + length > buf->maxsize)
	{
		//if (!buf->allowoverflow)
		//	Printf("SZ_GetSpace: overflow without allowoverflow set\n");
		
		if (length > buf->maxsize)
			Printf("SZ_GetSpace: %i is > full buffer size\n", length);
			
		Printf ("SZ_GetSpace: overflow\n");
		SZ_Clear (buf); 
		//buf->overflowed = true;
	}

	data = buf->data + buf->cursize;
	buf->cursize += length;
	
	return data;
}

void SZ_Write (sizebuf_t *buf, void *data, int length)
{
	memcpy (SZ_GetSpace(buf,length),data,length);		
}

void SZ_Write (sizebuf_t *b, byte *data, int startpos, int length)
{
    data += startpos;
    memcpy (SZ_GetSpace(b, length), data, length);
}


//
//
// ZDAEMON I/O PACKET FUNCTIONS
//
//

void ZD_WriteByte (sizebuf_t *sb, int c)
{
	byte	*buf;
	
	buf = SZ_GetSpace (sb, 1);
	buf[0] = c;
}

void ZD_WriteBytes(sizebuf_t *sb, byte *inbuf, int num)
{
	byte	*buf;
	
	buf = SZ_GetSpace (sb, num);
	memcpy(buf,inbuf,num);
}


void ZD_WriteWord (sizebuf_t *sb, short v)
{
	byte	*buf;

	buf = SZ_GetSpace(sb, 2);
	buf[0] = v&255;
	buf[1] = v>>8;
	
}

void ZD_WriteShort (sizebuf_t *sb, int c)
{
	byte	*buf;
	
	buf = SZ_GetSpace (sb, 2);
	buf[0] = c&0xff;
	buf[1] = c>>8;
}

void ZD_WriteLong (sizebuf_t *sb, int c)
{
	byte	*buf;
	
	buf = SZ_GetSpace (sb, 4);
	buf[0] = c&0xff;
	buf[1] = (c>>8)&0xff;
	buf[2] = (c>>16)&0xff;
	buf[3] = c>>24;
}

void ZD_WriteString (sizebuf_t *sb, char *s)
{
	if (!s)
		SZ_Write (sb, "", 1);
	else
		SZ_Write (sb, s, strlen(s)+1);
}


//
// reading functions
//

// returns -1 if no more characters are available

int ZD_ReadChar (void)
{
	int	c;
	
	if (msg_read.readcount+1 > msg_read.cursize)
		c = -1;
	else
		c = (signed char)msg_read.data[msg_read.readcount];
	msg_read.readcount++;
	
	return c;
}

int ZD_ReadByte (void)
{
	int	c;
	
	if (msg_read.readcount+1 > msg_read.cursize)
		c = -1;
	else
		c = (unsigned char)msg_read.data[msg_read.readcount];
	msg_read.readcount++;
	
	return c;
}

bool ZD_ReadBytes(byte *buf,int bufsz)
{
	if (msg_read.readcount+bufsz > msg_read.cursize)
	{
		memset(buf,0,bufsz);
		return false;
	}
	else
	{
		memcpy(buf,&(msg_read.data[msg_read.readcount]),bufsz);
		msg_read.readcount += bufsz;
		return true;
	}
}

int ZD_ReadShort (void)
{
	int	c;
	
	if (msg_read.readcount+2 > msg_read.cursize)
		c = -1;
	else		
		c = (short)(msg_read.data[msg_read.readcount]
		+ (msg_read.data[msg_read.readcount+1]<<8));
	
	msg_read.readcount += 2;
	
	return c;
}

int ZD_ReadWord (void)
{
	short v;// = (((*stream)[0]) << 8) | (((*stream)[1]));
	if (msg_read.readcount+2 > msg_read.cursize)
		v = -1;
	else
		v = (msg_read.data[msg_read.readcount]) 
			+ (msg_read.data[msg_read.readcount+1] << 8);
	
	msg_read.readcount += 2;
	return v;
}

int ZD_ReadLong (void)
{
	int	c;
	
	if (msg_read.readcount+4 > msg_read.cursize)
		c = -1;
	else
		c = msg_read.data[msg_read.readcount]
		+ (msg_read.data[msg_read.readcount+1]<<8)
		+ (msg_read.data[msg_read.readcount+2]<<16)
		+ (msg_read.data[msg_read.readcount+3]<<24);
	
	msg_read.readcount += 4;
	
	return c;
}

float ZD_ReadFloat (void)
{
	union
	{
		byte	b[4];
		float	f;
		int	l;
	} dat;
	
	if (msg_read.readcount+4 > msg_read.cursize)
		dat.f = -1;
	else
	{
		dat.b[0] =	msg_read.data[msg_read.readcount];
		dat.b[1] =	msg_read.data[msg_read.readcount+1];
		dat.b[2] =	msg_read.data[msg_read.readcount+2];
		dat.b[3] =	msg_read.data[msg_read.readcount+3];
	}
	msg_read.readcount += 4;
	
	//dat.l = LittleLong (dat.l);

	return dat.f;	
}

char *ZD_ReadString (void)
{
	static char	string[2048];
	int		l,c;
	
	l = 0;
	do
	{
		c = ZD_ReadChar ();
		if (c == -1 || c == 0)
			break;
		string[l] = c;
		l++;
	} while (l < sizeof(string)-1);
	
	string[l] = 0;
	
	return string;
}
