
//**************************************************************************
//**
//** ct_chat.c : Heretic 2 : Raven Software, Corp.
//**
//** $RCSfile: ct_chat.c,v $
//** $Revision: 1.12 $
//** $Date: 96/01/16 10:35:26 $
//** $Author: bgokey $
//**
//**************************************************************************

#include <string.h>
#include <ctype.h>
#include "doomdef.h"
#include "z_zone.h"
#include "m_swap.h"
#include "hu_stuff.h"
#include "w_wad.h"
#include "s_sound.h"
#include "doomstat.h"
#include "st_stuff.h"
#include "c_console.h"
#include "c_dispatch.h"
#include "c_cvars.h"
// [nightfang] - needed
#include "d_player.h"

#include "v_text.h"
#include "v_video.h"
#include "gi.h"
#include "d_gui.h"
#include "i_input.h"

// [NightFang] - client/server stuff
#include "cl_main.h"

#define QUEUESIZE		128
//#define MESSAGESIZE		128
//#define MESSAGELEN		265
#define HU_INPUTX		0
#define HU_INPUTY		(0 + (screen->Font->GetHeight () + 1))

EXTERN_CVAR (Bool, con_scaletext)
extern int	timeleft;		// in ticks

// Public data

void CT_Init ();
void CT_Drawer ();
BOOL CT_Responder (event_t *ev);
void HU_DrawScores (player_t *plyr);

int chatmodeon;

// Private data

static void CT_ClearChatMessage ();
static void CT_AddChar (char c);
static void CT_BackSpace ();
static void ShoveChatStr (const char *str, int who, int chat_type);

static int len;
static byte ChatQueue[QUEUESIZE];

CVAR (String, chatmacro1, "I'm ready to kick butt!", CVAR_ARCHIVE)
CVAR (String, chatmacro2, "I'm OK.", CVAR_ARCHIVE)
CVAR (String, chatmacro3, "I'm not looking too good!", CVAR_ARCHIVE)
CVAR (String, chatmacro4, "Help!", CVAR_ARCHIVE)
CVAR (String, chatmacro5, "You suck!", CVAR_ARCHIVE)
CVAR (String, chatmacro6, "Next time, scumbag...", CVAR_ARCHIVE)
CVAR (String, chatmacro7, "Come here!", CVAR_ARCHIVE)
CVAR (String, chatmacro8, "I'll take care of it.", CVAR_ARCHIVE)
CVAR (String, chatmacro9, "Yes", CVAR_ARCHIVE)
CVAR (String, chatmacro0, "No", CVAR_ARCHIVE)

FStringCVar *chat_macros[10] =
{
	&chatmacro0,
	&chatmacro1,
	&chatmacro2,
	&chatmacro3,
	&chatmacro4,
	&chatmacro5,
	&chatmacro6,
	&chatmacro7,
	&chatmacro8,
	&chatmacro9
};

EXTERN_CVAR (Bool, cl_rankspread)

//===========================================================================
//
// CT_Init
//
// 	Initialize chat mode data
//===========================================================================

void CT_Init ()
{
	len = 0; // initialize the queue index
	chatmodeon = 0;
	ChatQueue[0] = 0;
}

//===========================================================================
//
// CT_Stop
//
//===========================================================================

void CT_Stop ()
{
	chatmodeon = 0;
}

//===========================================================================
//
// CT_Responder
//
//===========================================================================

BOOL CT_Responder (event_t *ev)
{
	if (chatmodeon && ev->type == EV_GUI_Event)
	{
		if (ev->subtype == EV_GUI_KeyDown || ev->subtype == EV_GUI_KeyRepeat)
		{
			if (ev->data1 == '\r')
			{
				ShoveChatStr ((char *)ChatQueue, chatmodeon - 1, chatmodeon-1 );	//Kilgore
				CT_Stop ();
				return true;
			}
			else if (ev->data1 == GK_ESCAPE)
			{
				CT_Stop ();
				return true;
			}
			else if (ev->data1 == '\b')
			{
				CT_BackSpace ();
				return true;
			}
			else if (ev->data1 == 'C' && (ev->data3 & GKM_CTRL))
			{
				I_PutInClipboard ((char *)ChatQueue);
				return true;
			}
			else if (ev->data1 == 'V' && (ev->data3 & GKM_CTRL))
			{
				char *clip = I_GetFromClipboard ();
				if (clip != NULL)
				{
					char *clip_p = clip;
					strtok (clip, "\n\r\b");
					while (*clip_p)
					{
						CT_AddChar (*clip_p++);
					}
					delete[] clip;
				}
			}
		}
		else if (ev->subtype == EV_GUI_Char)
		{
			// send a macro
			if (ev->data2 && (ev->data1 >= '0' && ev->data1 <= '9'))
			{
				ShoveChatStr (*chat_macros[ev->data1 - '0'], chatmodeon-1, chatmodeon-1);
				CT_Stop ();
			}
			else
			{
				CT_AddChar (ev->data1);
			}
			return true;
		}
	}

	return false;
}

//===========================================================================
//
// CT_Drawer
//
//===========================================================================

void CT_Drawer (void)
{
	if (chatmodeon)
	{
		static const char *prompt = "Say: ";
		int i, x, scalex, y, promptwidth;

		if (con_scaletext)
		{
			scalex = CleanXfac;
			y = (!viewactive ? -30 : -10) * CleanYfac;
		}
		else
		{
			scalex = 1;
			y = (!viewactive ? -30 : -10);
		}

		y += (SCREENHEIGHT == realviewheight && viewactive) ? SCREENHEIGHT : ST_Y;

		promptwidth = screen->StringWidth (prompt) * scalex;
		x = screen->Font->GetCharWidth ('_') * scalex * 2 + promptwidth;

		// figure out if the text is wider than the screen->
		// if so, only draw the right-most portion of it.
		for (i = len - 1; i >= 0 && x < SCREENWIDTH; i--)
		{
			x += screen->Font->GetCharWidth (ChatQueue[i] & 0x7f) * scalex;
		}

		if (i >= 0)
		{
			i++;
		}
		else
		{
			i = 0;
		}

		// draw the prompt, text, and cursor
		ChatQueue[len] = gameinfo.gametype == GAME_Doom ? '_' : '[';
		ChatQueue[len+1] = '\0';
		if (con_scaletext)
		{
			screen->DrawTextClean (CR_GREEN, 0, y, prompt);
			screen->DrawTextClean (CR_GREY, promptwidth, y, ChatQueue + i);
		}
		else
		{
			screen->DrawText (CR_GREEN, 0, y, prompt);
			screen->DrawText (CR_GREY, promptwidth, y, ChatQueue + i);
		}
		ChatQueue[len] = '\0';

		BorderTopRefresh = screen->GetPageCount ();
	}

	// [NightFang] - it dosent have to be deathmatch anymore
	//if (deathmatch &&
	if	(Button_ShowScores.bDown ||
		 players[consoleplayer].camera->health <= 0)
	{
		HU_DrawScores (&players[consoleplayer]);
	}
}

//===========================================================================
//
// CT_AddChar
//
//===========================================================================

static void CT_AddChar (char c)
{
	if (len < QUEUESIZE-2)
	{
		ChatQueue[len++] = c;
		ChatQueue[len] = 0;
	}
}

//===========================================================================
//
// CT_BackSpace
//
// 	Backs up a space, when the user hits (obviously) backspace
//===========================================================================

static void CT_BackSpace ()
{
	if (len)
	{
		ChatQueue[--len] = 0;
	}
}

//===========================================================================
//
// CT_ClearChatMessage
//
// 	Clears out the data for the chat message.
//===========================================================================

static void CT_ClearChatMessage ()
{
	ChatQueue[0] = 0;
	len = 0;
}

/******************************************************
 * Kilgore: support for chat spam control.
 ******************************************************/
EXTERN_CVAR (Int, spam_window)
EXTERN_CVAR (Int, spam_limit)

#define	CHAT_SPAM_QUEUESZ		64					// Must be more than "spam_limit".
													// We check for repeated msgs in
													// the last CHAT_SPAM_QUEUESZ msgs
													// that occurred in the spam control
													// interval.

class ChatSpamQueue
{
private:
	unsigned	front, rear, nitems;
	int			msg_time[CHAT_SPAM_QUEUESZ];
	char		*msg_text[CHAT_SPAM_QUEUESZ];
	void		init();

public:
				ChatSpamQueue();
				~ChatSpamQueue();
	bool		is_empty()	{ return (nitems==0); }
	bool		is_full()	{ return (nitems==CHAT_SPAM_QUEUESZ); }
	int			length()	{ return nitems; }
	void		flush();
	bool		add(const char *msg);
	void		trim(void);
	bool		in(const char *s);
};

ChatSpamQueue::ChatSpamQueue()
{
	init();
}

ChatSpamQueue::~ChatSpamQueue()
{
	flush();
}

void ChatSpamQueue::init()
{
	front = rear = nitems = 0;
	memset(msg_text,0,sizeof(msg_text));
	memset(msg_time,0,sizeof(msg_time));
}

void ChatSpamQueue::flush()
{
	unsigned i;

	for (i=front;  nitems>0;  nitems--)
	{
		free(msg_text[i]);
		i = (i+1) % CHAT_SPAM_QUEUESZ;
	}
	init();
}

bool ChatSpamQueue::add(const char *mtext)
{
	char *s;

	if (is_full()) return false;
	if (mtext==NULL || (s=strdup(mtext))==NULL ) return false;
	msg_text[rear] = s;
	msg_time[rear] = gametic;
	rear = (rear + 1) % CHAT_SPAM_QUEUESZ;
	nitems++;
	return true;
}

void ChatSpamQueue::trim()
{
	unsigned	i;
	int			cutoff;

	cutoff = gametic - TICRATE * (int)spam_window;
	for (i=front;  nitems>0;   nitems--)
	{
		if (msg_time[i]>cutoff) break;
		free(msg_text[i]);
		msg_text[i] = NULL;
		msg_time[i] = 0;
		i = (i+1) % CHAT_SPAM_QUEUESZ;
	}
	front = i;
}

bool ChatSpamQueue::in(const char *s)
{
	unsigned	i,n;

	i = front;
	for (n=nitems; n>0;  n--)
	{
		if (!stricmp(s,msg_text[i])) return true;
		i = (i+1) % CHAT_SPAM_QUEUESZ;
	}
	return false;
}

static ChatSpamQueue chat_spam_queue;

/******************************************************/

static void ShoveChatStr(const char *str, int who, int chat_type)
{

	// [NightFang] - the old way if we're playing locally
	if (zd_localplay)
	{
		Net_WriteByte (DEM_SAY);
		Net_WriteByte ((byte)who);
		Net_WriteString (str);
		return;
	}

	// Spectators should not send team or private chat messages; cheating opportunity
	if (chat_type!=0 && players[consoleplayer].playerstate==PST_WATCHING)
	{
		Printf("Spectators cannot send team or private chat messages\n");
		return;
	}

	// Players should not send private chat messages to members of other teams
	// in teamplay mode; cheating opportunity
	if (chat_type==2 && teamplay && players[who].userinfo.team != players[consoleplayer].userinfo.team)
	{
		Printf("You cannot send a private chat message to members of opposing teams\n");
		return;
	}

	/*******************************************************
	 * Kilgore: Chat spam control
	 *******************************************************/
	if (spam_limit>0 && spam_window>0)
	{
		const char *pin;
		char *pout,temp[512];

		// Copy msg, ignoring blanks and periods.
		pout = temp;
		for (pin=str; *pin;  pin++)
		{
			if (*pin!=' ' && *pin!='.') *pout++ = *pin;
		}
		*pout = '\0';

		chat_spam_queue.trim();					//Get rid of older messages
		if (temp[0]=='\0')
		{
			chat_spam_queue.add("");			//Hit him with a penalty delay
			Printf("Any other enlightening remarks?\n");
			return;
		}
		if (chat_spam_queue.in(temp))
		{
			chat_spam_queue.add("");			//Hit him with a penalty delay
			Printf("You've just said that!\n");
			return;
		}
		if ( chat_spam_queue.length() >= (int)spam_limit)
		{
			chat_spam_queue.add("");			//Hit him with a penalty delay
			Printf("Talk is cheap. Get up and fight soldier!\n");
			return;
		}
		chat_spam_queue.add(temp);				//Accept it
	}

	/*******************************************************/

	// [NightFang] - we'll just do this the zdaemon way
	switch (chat_type)
	{
		case 0:
			ZD_WriteByte(&localbuf,cl_chat);
			ZD_WriteString(&localbuf, (char *) str);
			break;
		case 1:
			ZD_WriteByte(&localbuf,cl_teamchat);
			ZD_WriteString(&localbuf, (char *) str);
			break;
		case 2:
			ZD_WriteByte(&localbuf,cl_privchat);
			ZD_WriteByte(&localbuf,who);
			ZD_WriteString(&localbuf, (char *) str);
			break;
	}
}

CCMD (messagemode)
{
	if (!menuactive)
	{
		chatmodeon = 1;
		C_HideConsole ();
		CT_ClearChatMessage ();
	}
}

static char *collect_args(FCommandLine &argv, int firsti)
{
	int				i,n;
	static char		buf[512];

	n = argv.argc();
	buf[0] = '\0';
	for (i=firsti; i<n; i++)
	{
		if (i>1) strcat(buf," ");
		strcat(buf,argv[i]);
	}
	return buf;
}

CCMD (say)
{
	ShoveChatStr(collect_args(argv,1), 0, 0);
}

CCMD (messagemode2)
{
	if (!menuactive)
	{
		chatmodeon = 2;
		C_HideConsole ();
		CT_ClearChatMessage ();
	}
}

CCMD (say_team)
{
	ShoveChatStr(collect_args(argv,1), 0, 1);
}

CCMD (sayto)
{
	int		i, j;
	char	*p;

	if (argv.argc() < 3) 
	{
		USAGE:
		Printf(	"Usage:\n"
			"   sayto $player_name <message>   or\n"
			"   sayto #player_id   <message>\n");
		return;
	}
	p = argv[1];
	if (*p=='#')
	{
		p++;
		if (sscanf(p,"%d",&i)!=1 || i<0 || i>=MAXPLAYERS || !playeringame[i])
			goto USAGE;
		else
			ShoveChatStr(collect_args(argv,2), i, 2);
	}
	else if (*p=='$')
	{
		p++;
		if (*p=='\0') goto USAGE;
		for (j=MAXPLAYERS-1; j>=0; j--)
		{
			if (!playeringame[j]) continue;
			if (!stricmp(p,players[j].userinfo.netname)) break;
		}
		if (j<0)	Printf("Unknown player name\n");
		else		ShoveChatStr(collect_args(argv,2), j, 2);
	}
	else
	{
		goto USAGE;
	}
}

static int STACK_ARGS compare_coopplayers(const void *arg1, const void *arg2)
{
	return (* (player_t **) arg2)->killcount - (* (player_t **) arg1)->killcount;
}

static int STACK_ARGS compare_players(const void *arg1, const void *arg2)
{
	int rc;
	
	rc = (* (player_t **) arg2)->fragcount - (* (player_t **) arg1)->fragcount;
	if (rc==0)
		rc = (* (player_t **) arg1)->deathcount - (* (player_t **) arg2)->deathcount;
	return rc;
}

static int STACK_ARGS compare_teams(const void *arg1, const void *arg2)
{
	return (* (team_info_s **) arg2)->score - (* (team_info_s **) arg1)->score;
}

EXTERN_CVAR (Int, timelimit)
EXTERN_CVAR (Int, fraglimit)
EXTERN_CVAR (Int, teamscorelimit)

// Lyfe - Jan 10, 2005
// I wanted to cut down on the text in the functions below, when all that's needed is a check
// for whether or not the text should be translucent.
inline void scoreboardDrawFunction(bool translucent, int a, int b, int c, const char *text)
{
	if (translucent)
		screen->DrawTextCleanLuc(a, b, c, text);
	else
		screen->DrawTextClean(a, b, c, text);
}

static int max_timer_width()
{
	int w1,w2,maxw;

	w1 = screen->StringWidth("88:88");
	w2 = screen->StringWidth("00:00");
	maxw = (w1>w2) ? w1 : w2;
	return maxw * CleanXfac;
}

static char *timer_string()
{
	int			nsecs, nmins;
	static int	last_nsecs;
	static char	str[64];

	nsecs = (timeleft+TICRATE-1)/TICRATE;
	if (last_nsecs==nsecs-1 || last_nsecs==nsecs-2)
		nsecs = last_nsecs;
	else
		last_nsecs = nsecs;
	nmins = nsecs / 60;
	nsecs -= nmins * 60;
	sprintf(str,"%02d:%02d",nmins,nsecs);
	return str;
}

//Kilgore: Team DM score page
static void Draw_TeamDMScores(player_t *player)
{
	int				i, j, y, nxscreen, nyscreen, xt[NUM_TEAMS], yt[NUM_TEAMS], yspace,
					maxyt[NUM_TEAMS], membercount[NUM_TEAMS], text_height, maxv, lim1, lim2;
	int				tm, width, namewidth[2], maxwidth[2], color, nxborder, ht1, ht2;
	team_info_s		*ti;
	player_t		*pi, *sortedplayers[MAXPLAYERS];
	char			str[256];


	if (player->camera->player)
		player = player->camera->player;

	sortedplayers[MAXPLAYERS-1] = player;
	for (i = 0, j = 0; j < MAXPLAYERS - 1; i++, j++)
	{
		if (&players[i] == player)
			i++;
		sortedplayers[j] = &players[i];
	}

	qsort (sortedplayers, MAXPLAYERS, sizeof(player_t *), compare_players);

screen->SetFont(ConFont);

	memset(membercount,0,sizeof(membercount));
	memset(namewidth,0,sizeof(namewidth));
	for (i=0; i<NUM_TEAMS; i++)
	{
		width = screen->StringWidth(TeamInfo[i].name);
		j = i & 1;
		if (width > namewidth[j]) namewidth[j] = width;
	}
	for (i=0; i<MAXPLAYERS; i++)
	{
		if (!playeringame[i]) continue;
		tm = players[i].userinfo.team;
		if (tm<0 || tm>=NUM_TEAMS) tm = NUM_TEAMS - 1;
		(membercount[tm])++;
		j = tm & 1;
		width = screen->StringWidth (players[i].userinfo.netname);
		if (width > namewidth[j]) namewidth[j] = width;
	}
	for (i=0; i<2; i++)
	{
		namewidth[i] += screen->StringWidth(" ");
		maxwidth[i] = namewidth[i] + screen->StringWidth("99/9999");
		namewidth[i] *= CleanXfac;
		maxwidth[i] *= CleanXfac;
	}

	nxscreen = SCREENWIDTH;

	// If possible, try to make the widths equal
	maxv = (maxwidth[0] > maxwidth[1]) ? maxwidth[0] : maxwidth[1];
	if (2*maxv+6<=nxscreen)
	{
		maxwidth[0] = maxwidth[1] = maxv;
		if (namewidth[0]>namewidth[1])
			namewidth[1] = namewidth[0];
		else
			namewidth[0] = namewidth[1];
	}

	nxborder = (nxscreen-maxwidth[0]-maxwidth[1]) / 3;
	if (nxborder<0) nxborder = 0;
	xt[0] = xt[2] = nxborder;
	xt[1] = xt[3] = nxborder + maxwidth[0] + nxborder;

	text_height = (screen->Font->GetHeight()+1) * CleanYfac;
	nyscreen = SCREENHEIGHT;
	y = 2*CleanYfac;

	if (consoleplayer==displayplayer)
	{
		sprintf(str, "Exp: %d", players[consoleplayer].experience);
		screen->DrawTextClean(CR_GRAY, nxscreen-screen->StringWidth(str) * CleanXfac, y, str);
	}
	y += text_height + CleanYfac;

	if (timeleft>0)
		screen->DrawTextClean(CR_GRAY, nxscreen - max_timer_width(), y, timer_string());

	strcpy(str,"Team  Deathmatch");
	screen->DrawTextClean (CR_GOLD, (nxscreen - screen->StringWidth(str)*CleanXfac)/2,y,str);
	y += text_height+CleanYfac;

	lim1 = (int) fraglimit;
	lim2 = (int) teamscorelimit;
	if (lim1>0 && lim2>0)	sprintf(str,"Frag Limit: %d    Team Limit: %d",lim1, lim2);
	else if (lim1>0)		sprintf(str,"Frag Limit: %d",lim1);
	else if (lim2>0)		sprintf(str,"Team Score Limit: %d",lim2);
	else					str[0] = '\0';
	if (str[0])
	{
		screen->DrawTextClean(CR_WHITE, (nxscreen - screen->StringWidth(str)*CleanXfac)/2,y,str);
		y += text_height+CleanYfac;
	}

	// Get the required height for each row
	maxv = (membercount[0]>membercount[1]) ? membercount[0] : membercount[1];
	ht1 = (maxv+1)*text_height + 	4*CleanYfac;
	maxv = (membercount[2]>membercount[3]) ? membercount[2] : membercount[3];
	ht2 = (maxv+1)*text_height + 	4*CleanYfac;

	// If possible, try to make them equal
	maxv = (ht1>ht2) ? ht1 : ht2;
	if (2*maxv+y+14<=nyscreen) ht1 = ht2 = maxv;

	yspace = (nyscreen-y-ht1-ht2)/7;
	if (yspace<0) yspace = 0;

	yt[0] = yt[1] = y + yspace;
	yt[2] = yt[3] = yt[0] + ht1 + 2*yspace;

	maxyt[0] = maxyt[1] = yt[2] - text_height;
	maxyt[2] = maxyt[3] = nyscreen - text_height;

	for (i=0; i<NUM_TEAMS; i++)
	{
		ti = TeamInfo + i;
		screen->DrawTextClean(ti->color, xt[i], yt[i], TeamInfo[i].name);
		sprintf (str, "%d", ti->score);
		screen->DrawTextClean(ti->color, xt[i]+namewidth[i & 1], yt[i], str);
		yt[i] += text_height + 4*CleanYfac;
	}

	for (i=0; i<MAXPLAYERS; i++)
	{
		pi = sortedplayers[i];

		if (!playeringame[pi - players]) continue;
		tm = pi->userinfo.team;
		if (tm<0 || tm>=NUM_TEAMS)
		{
			tm = NUM_TEAMS-1;
			color = CR_WHITE;		//Visual indication that something went wrong
		}
		else
		{
			color = (pi==player) ?									CR_BRICK :
					(demoplayback && pi==&players[consoleplayer]) ?	CR_BRICK :
																	TeamInfo[tm].color;
		}
		y = yt[tm];
		if (y>=maxyt[tm]) continue;

		if (pi->isbot)	sprintf(str,"%d/-",pi->fragcount);
		else			sprintf(str,"%d/%d",pi->fragcount,pi->ping);

		if (pi->playerstate == PST_WATCHING)
		{
			screen->DrawTextCleanLuc(color,xt[tm],y,pi->userinfo.netname);
			screen->DrawTextCleanLuc(color,xt[tm]+namewidth[tm & 1],y,str);
		}
		else
		{
			screen->DrawTextClean(color,xt[tm],y,pi->userinfo.netname);
			screen->DrawTextClean(color,xt[tm]+namewidth[tm & 1],y,str);
		}

		yt[tm] +=  text_height;
	}
screen->SetFont(SmallFont);
}

#if __CTF__		//CTF_FIX
#if 0
//duke
//==
static void Draw_CTFScores(player_t *player)
{
	int				i, j, y, nxscreen, nyscreen, xt[NUM_TEAMS], yt[NUM_TEAMS], yspace,
					maxyt[NUM_TEAMS], membercount[NUM_TEAMS], text_height, maxv, lh,
					tm, width, namewidth[2], maxwidth[2], color, nxborder, ht1,ht2;
	team_info_s		*ti;
	patch_t			*logo;
	player_t		*pi, *sortedplayers[MAXPLAYERS];
	char			str[256];

	if (player->camera->player)
		player = player->camera->player;

	sortedplayers[MAXPLAYERS-1] = player;
	for (i = 0, j = 0; j < MAXPLAYERS - 1; i++, j++)
	{
		if (&players[i] == player)
			i++;
		sortedplayers[j] = &players[i];
	}

	qsort (sortedplayers, MAXPLAYERS, sizeof(player_t *), compare_players);

	memset(membercount,0,sizeof(membercount));
	memset(namewidth,0,sizeof(namewidth));
	for (i=0; i<NUM_TEAMS; i++)
	{
		width = screen->StringWidth(TeamInfo[i].name);
		j = i & 1;
		if (width > namewidth[j]) namewidth[j] = width;
	}
	for (i=0; i<MAXPLAYERS; i++)
	{
		if (!playeringame[i]) continue;
		tm = players[i].userinfo.team;
		if (tm<0 || tm>=NUM_TEAMS) tm = NUM_TEAMS - 1;
		(membercount[tm])++;
		j = tm & 1;
		width = screen->StringWidth (players[i].userinfo.netname);
		if (width > namewidth[j]) namewidth[j] = width;
	}
	for (i=0; i<2; i++)
	{
		namewidth[i] += screen->StringWidth(" ");
		maxwidth[i] = namewidth[i] + screen->StringWidth("99/9999");
		namewidth[i] *= CleanXfac;
		maxwidth[i] *= CleanXfac;
	}

	nxscreen = SCREENWIDTH;

	// If possible, try to make the widths equal
	maxv = (maxwidth[0] > maxwidth[1]) ? maxwidth[0] : maxwidth[1];
	if (2*maxv+6<=nxscreen) maxwidth[0] = maxwidth[1] = maxv;

	nxborder = (nxscreen-maxwidth[0]-maxwidth[1]) / 3;
	if (nxborder<0) nxborder = 0;
	xt[0] = xt[2] = nxborder;
	xt[1] = xt[3] = nxborder + maxwidth[0] + nxborder;

	text_height = (screen->Font->GetHeight()+1) * CleanYfac;
	nyscreen = SCREENHEIGHT;
	y = 2*CleanYfac;

	if (consoleplayer==displayplayer)
	{
		sprintf(str, "Exp: %d", players[consoleplayer].experience);
		screen->DrawTextClean(CR_GRAY, nxscreen-screen->StringWidth(str) * CleanXfac, y, str);
	}
	if (timeleft>0)
		screen->DrawTextClean(CR_GRAY, (nxscreen - max_timer_width())/2,y,timer_string());
	y += text_height + CleanYfac;

	strcpy(str,"Capture the Flag");
	screen->DrawTextClean (CR_GOLD, (nxscreen - screen->StringWidth(str)*CleanXfac)/2,y,str);
	y += text_height+CleanYfac;

	if (teamscorelimit>0)
	{
		sprintf(str,"Score Limit: %d",(int)teamscorelimit);
		screen->DrawTextClean(CR_WHITE, (nxscreen - screen->StringWidth(str)*CleanXfac)/2,y,str);
		y += text_height+CleanYfac;
	}

	y += 4*CleanYfac;
	for (i=0; i<2; i++)
	{
		sprintf(str,"ctfl%d",i);
		logo = (patch_t *)W_CacheLumpName(str,101);
		lh = logo->height*CleanYfac;
		screen->DrawPatchStretched(logo,xt[i],y, logo->width*CleanXfac,lh);

		ti = TeamInfo + i;

		screen->SetFont(BigFont);
		sprintf (str, "%d", ti->score);
		screen->DrawTextClean(ti->color, xt[i]+namewidth[i & 1], y+lh/4, str);
		screen->SetFont(SmallFont);
	}
	y += lh;
	y -= 4*CleanYfac;

	// Get the required height for each row
	maxv = (membercount[0]>membercount[1]) ? membercount[0] : membercount[1];
	ht1 = maxv*text_height + 4*CleanYfac;
	maxv = (membercount[2]>membercount[3]) ? membercount[2] : membercount[3];
	ht2 = maxv*text_height + 4*CleanYfac;

	// If possible, try to make them equal
	maxv = (ht1>ht2) ? ht1 : ht2;
	if (2*maxv+y+14<=nyscreen) ht1 = ht2 = maxv;

	yspace = (nyscreen-y-ht1-ht2)/7;
	if (yspace<0) yspace = 0;

	yt[0] = yt[1] = y + yspace;
	yt[2] = yt[3] = yt[0] + ht1 + 2*yspace;

	maxyt[0] = maxyt[1] = yt[2] - text_height;
	maxyt[2] = maxyt[3] = nyscreen - text_height;

	for (i=0; i<MAXPLAYERS; i++)
	{
		pi = sortedplayers[i];

		if (!playeringame[pi - players]) continue;
		tm = pi->userinfo.team;
		if (tm<0 || tm>=NUM_TEAMS)
		{
			tm = NUM_TEAMS-1;
			color = CR_WHITE;		//Visual indication that something went wrong
		}
		else if (tm>=2)
		{
			color = (pi==player) ?									CR_BRICK :
					(demoplayback && pi==&players[consoleplayer]) ?	CR_BRICK :
																	CR_GRAY;
		}
		else
		{
			color = (pi==player) ?									CR_BRICK :
					(demoplayback && pi==&players[consoleplayer]) ?	CR_BRICK :
																	TeamInfo[tm].color;
		}
		y = yt[tm];
		if (y>=maxyt[tm]) continue;

		if (pi->isbot)	sprintf(str,"%d/-",pi->fragcount);
		else			sprintf(str,"%d/%d",pi->fragcount,pi->ping);

		if (pi->playerstate == PST_WATCHING)
		{
			screen->DrawTextCleanLuc(color,xt[tm],y,pi->userinfo.netname);
			screen->DrawTextCleanLuc(color,xt[tm]+namewidth[tm & 1],y,str);
		}
		else
		{
			screen->DrawTextClean(color,xt[tm],y,pi->userinfo.netname);
			screen->DrawTextClean(color,xt[tm]+namewidth[tm & 1],y,str);
		}

		yt[tm] +=  text_height;
	}
}
//== end

#elif 1

static void Draw_CTFScores(player_t *player)
{
	int				i, j, y, nxscreen, nyscreen,
					membercount[NUM_TEAMS], text_height,
					namewidth[2], color;
	//team_info_s		*ti;
	//patch_t			*logo;
	player_t		*pi, *sortedplayers[MAXPLAYERS];
	char			str[256];
	extern int ctf_carrier_id[2];

	byte teamcolor[2];
	teamcolor[0] = 185; // doom palette: red
	teamcolor[1] = 201; // doom palette: blue
	// gonna put the players in a box. :)
	int redboxLeft,  redboxRight,  redboxTop,  redboxBottom;
	int blueboxLeft, blueboxRight, blueboxTop, blueboxBottom;
	fixed_t translucency[2] = { 0x4000, 0x4000 };
	int ypos[3];
	int xpos[3];
	int teamnum;
	int offsetName;//  = ( 1.0/100.0) * nxscreen;
	int offsetFrags;// = (50.0/100.0) * nxscreen;
	int offsetPing;//  = (70.0/100.0) * nxscreen;
	int flagsizex, flagsizey;
	bool translucent;
	player_s *redcarrier  = players + ctf_carrier_id[0];
	player_s *bluecarrier = players + ctf_carrier_id[1];
	
	if (player->camera->player)
		player = player->camera->player;

	// sort players
	sortedplayers[MAXPLAYERS-1] = player;
	for (i = 0, j = 0; j < MAXPLAYERS - 1; i++, j++) {
		if (&players[i] == player) {
			i++;
		}
		sortedplayers[j] = &players[i];
	}
	qsort (sortedplayers, MAXPLAYERS, sizeof(player_t *), compare_players);
	
	screen->SetFont(ConFont);
	// determine team sizes.
	memset(membercount,0,sizeof(membercount));
	memset(namewidth,0,sizeof(namewidth));
	for (i=0; i<MAXPLAYERS; i++)
	{
		if (!playeringame[i])
			continue;
		teamnum = players[i].userinfo.team;
		if (teamnum < 0 || teamnum >= NUM_TEAMS)
			teamnum = NUM_TEAMS - 1;
		(membercount[teamnum])++;
	}
	text_height = (screen->Font->GetHeight()+1) * CleanYfac;
	nyscreen = SCREENHEIGHT;
	nxscreen = SCREENWIDTH;
	
	// gonna use y to space out the top of the display.

	if (timeleft>0)
		screen->DrawTextClean(CR_GRAY, nxscreen - max_timer_width(), 0, timer_string());

	strcpy(str,"Capture the Flag");
	screen->DrawTextClean (CR_GOLD, (nxscreen - screen->StringWidth(str)*CleanXfac)/2, 0, str);

	if (teamscorelimit>0)
	{
		sprintf(str,"Score Limit: %d",(int)teamscorelimit);
		screen->DrawTextClean(CR_WHITE, (nxscreen - screen->StringWidth(str)*CleanXfac)/2, text_height * 1, str);
	}

	//y += 4*CleanYfac;
	redboxLeft    = ( 2.0/100.0) * nxscreen;
	redboxRight   = nxscreen - ( 2.0/100.0) * nxscreen;
	// start 3 lines down.
	redboxTop     = text_height * 3 + ( 2.0/100.0) * nyscreen;
	if(membercount[0] < 2) { // we really want the boxes to fit the team score right, which needs two rows.
		redboxBottom  = redboxTop + text_height * 2;
	} else {
		redboxBottom  = redboxTop + text_height * membercount[0];
	}
	blueboxLeft   = ( 2.0/100.0) * nxscreen;
	blueboxRight  = nxscreen - ( 2.0/100.0) * nxscreen;
	blueboxTop    = redboxBottom + (2.0/100) * nyscreen;
	if(membercount[1] < 2) {
		blueboxBottom = blueboxTop + text_height * 2;
	} else {
		blueboxBottom = blueboxTop + text_height * membercount[1];
	}
	ypos[0] =  redboxTop+1;
	ypos[1] = blueboxTop+1;
	ypos[2] = blueboxBottom + ( 2.0/100.0) * nyscreen;
	xpos[0] =  redboxLeft;
	xpos[1] = blueboxLeft;
	xpos[2] = blueboxLeft;
	// adding string width of "00" to this to make room for a flag icon.
	offsetName  = ( 1.0/100.0) * nxscreen + text_height; // + screen->StringWidth("Name");
	offsetFrags = (50.0/100.0) * nxscreen + screen->StringWidth("Score")*CleanXfac;
	offsetPing  = (70.0/100.0) * nxscreen + screen->StringWidth("Ping")*CleanXfac;
	flagsizex   = text_height;
	flagsizey   = text_height;

	// column headers.
	sprintf(str, "Name");
	screen->DrawTextClean(CR_WHITE, redboxLeft + offsetName, redboxTop - text_height, str);
	sprintf(str, "Points");
	screen->DrawTextClean(CR_WHITE, redboxLeft + offsetFrags - (screen->StringWidth(str)*CleanXfac), redboxTop - text_height, str);
	sprintf(str, "Ping");
	screen->DrawTextClean(CR_WHITE, redboxLeft + offsetPing  - (screen->StringWidth(str)*CleanXfac), redboxTop - text_height, str);
	sprintf(str, "Score");
	screen->DrawTextClean(CR_WHITE, redboxRight - (screen->StringWidth(str)*CleanXfac), redboxTop - text_height, str);

	// team boxes & scores, red & blue.
	screen->ScaleTranslucentMaskedBlock( redboxLeft, redboxTop,  1, 1,  redboxRight- redboxLeft,  redboxBottom- redboxTop, &(teamcolor[0]), NULL, translucency[0]);
	screen->ScaleTranslucentMaskedBlock(blueboxLeft, blueboxTop, 1, 1, blueboxRight-blueboxLeft, blueboxBottom-blueboxTop, &(teamcolor[1]), NULL, translucency[1]);
	screen->SetFont(BigFont);
	sprintf (str, "%d", TeamInfo[0].score);
	screen->DrawTextClean(TeamInfo[0].color,  redboxRight - screen->StringWidth(str)*CleanXfac,  redboxTop+1, str);
	sprintf (str, "%d", TeamInfo[1].score);
	screen->DrawTextClean(TeamInfo[1].color, blueboxRight - screen->StringWidth(str)*CleanXfac, blueboxTop+1, str);
	screen->SetFont(ConFont);
	
	// print player info.
	for (i=0; i<MAXPLAYERS; i++) {
		pi = sortedplayers[i];

		if (!playeringame[pi - players])
			continue;
		teamnum = pi->userinfo.team;
		// select color
		if (teamnum < 0 || teamnum >= NUM_TEAMS) {
			teamnum = NUM_TEAMS-1;
			color = CR_WHITE;		//Visual indication that something went wrong
		} else if (teamnum >= 2) {
			teamnum = 2;
			color = CR_GRAY;
			if(pi == player || (demoplayback && pi == &players[consoleplayer])) {
				color = CR_BRICK;
			}
		} else {
			color = TeamInfo[teamnum].color;
			if(pi == player || (demoplayback && pi == &players[consoleplayer])) {
				color = CR_BRICK;
			}
		}
		// figure out height position.
		y = ypos[teamnum];

		translucent = (pi->playerstate == PST_WATCHING);
		// display flag next to carrier
		if (pi == redcarrier)
			screen->DrawPatchStretched((patch_t *)W_CacheLumpName("ctf11",101), xpos[teamnum], y, flagsizex, flagsizey);
		if (pi == bluecarrier)
			screen->DrawPatchStretched((patch_t *)W_CacheLumpName("ctf21",101), xpos[teamnum], y, flagsizex, flagsizey);

		// print name.
		scoreboardDrawFunction(translucent, color, xpos[teamnum]+offsetName,  y, pi->userinfo.netname);

		// print frags
		sprintf(str, "%d", pi->fragcount);
		scoreboardDrawFunction(translucent, color, xpos[teamnum]+offsetFrags - screen->StringWidth(str)*CleanXfac, y, str);

		// print ping.
		if (pi->isbot)
			sprintf(str,"-");
		else
			sprintf(str,"%d",pi->ping);

		scoreboardDrawFunction(translucent, color, xpos[teamnum]+offsetPing - screen->StringWidth(str)*CleanXfac,  y, str);
		
		// increment the yposition for this team to the next row.
		ypos[teamnum] +=  text_height;
	}
	screen->SetFont(SmallFont);
}
#else

static void Draw_CTFScores(player_t *player)
{
	int				i, j, y, nxscreen, nyscreen, xt[NUM_TEAMS], yt[NUM_TEAMS], yspace,
					maxyt[NUM_TEAMS], membercount[NUM_TEAMS], text_height, maxv, 
					tm, width, namewidth[2], maxwidth[2], color, nxborder, ht1,ht2;
	team_info_s		*ti;
	player_t		*pi, *sortedplayers[MAXPLAYERS];
	char			str[256];

	if (player->camera->player)
		player = player->camera->player;

	sortedplayers[MAXPLAYERS-1] = player;
	for (i = 0, j = 0; j < MAXPLAYERS - 1; i++, j++)
	{
		if (&players[i] == player)
			i++;
		sortedplayers[j] = &players[i];
	}

	screen->SetFont(ConFont);

	qsort (sortedplayers, MAXPLAYERS, sizeof(player_t *), compare_players);

	memset(membercount,0,sizeof(membercount));
	memset(namewidth,0,sizeof(namewidth));
	for (i=0; i<NUM_TEAMS; i++)
	{
		width = screen->StringWidth(TeamInfo[i].name);
		j = i & 1;
		if (width > namewidth[j]) namewidth[j] = width;
	}
	for (i=0; i<MAXPLAYERS; i++)
	{
		if (!playeringame[i]) continue;
		tm = players[i].userinfo.team;
		if (tm<0 || tm>=NUM_TEAMS) tm = NUM_TEAMS - 1;
		(membercount[tm])++;
		j = tm & 1;
		width = screen->StringWidth (players[i].userinfo.netname);
		if (width > namewidth[j]) namewidth[j] = width;
	}
	for (i=0; i<2; i++)
	{
		namewidth[i] += screen->StringWidth(" ");
		maxwidth[i]   = namewidth[i] + screen->StringWidth("99/9999");
		namewidth[i] *= CleanXfac;
		maxwidth[i]  *= CleanXfac;
	}

	nxscreen = SCREENWIDTH;

	// If possible, try to make the widths equal
	maxv = (maxwidth[0] > maxwidth[1]) ? maxwidth[0] : maxwidth[1];
	if (2*maxv+6<=nxscreen)
	{
		maxwidth[0] = maxwidth[1] = maxv;
		if (namewidth[0]>namewidth[1])
			namewidth[1] = namewidth[0];
		else
			namewidth[0] = namewidth[1];
	}

	nxborder = (nxscreen-maxwidth[0]-maxwidth[1]) / 3;
	if (nxborder<0) nxborder = 0;
	xt[0] = xt[2] = nxborder;
	xt[1] = xt[3] = nxborder + maxwidth[0] + nxborder;

	text_height = (screen->Font->GetHeight()+1) * CleanYfac;
	nyscreen = SCREENHEIGHT;
	y = 2*CleanYfac;

	if (timeleft>0)
		screen->DrawTextClean(CR_GRAY, nxscreen - max_timer_width(),y,timer_string());
	y += text_height + CleanYfac;

	strcpy(str,"Capture the Flag");
	screen->DrawTextClean (CR_GOLD, (nxscreen - screen->StringWidth(str)*CleanXfac)/2,y,str);
	y += text_height+CleanYfac;

	if (teamscorelimit>0)
	{
		sprintf(str,"Score Limit: %d",(int)teamscorelimit);
		screen->DrawTextClean(CR_WHITE, (nxscreen - screen->StringWidth(str)*CleanXfac)/2,y,str);
		y += text_height+CleanYfac;
	}

	y += 4*CleanYfac;
	for (i=0; i<2; i++)
	{
		ti = TeamInfo + i;
		screen->DrawTextClean(ti->color, xt[i], y, ti->name);
		sprintf (str, "%d", ti->score);
		screen->DrawTextClean(ti->color, xt[i]+namewidth[i], y, str);

		byte jnk_data[256];
		memset(jnk_data,0,sizeof(jnk_data));

		screen->DrawBlock(xt[i],y+text_height, maxwidth[i], 3, jnk_data);
	}
	y += text_height;

	// Get the required height for each row
	maxv = (membercount[0]>membercount[1]) ? membercount[0] : membercount[1];
	ht1 = maxv*text_height + 4*CleanYfac;
	maxv = (membercount[2]>membercount[3]) ? membercount[2] : membercount[3];
	ht2 = maxv*text_height + 4*CleanYfac;

	// If possible, try to make them equal
	maxv = (ht1>ht2) ? ht1 : ht2;
	if (2*maxv+y+14<=nyscreen) ht1 = ht2 = maxv;

	yspace = (nyscreen-y-ht1-ht2)/7;
	if (yspace<0) yspace = 0;

	yt[0] = yt[1] = y + yspace;
	yt[2] = yt[3] = yt[0] + ht1 + 2*yspace;

	maxyt[0] = maxyt[1] = yt[2] - text_height;
	maxyt[2] = maxyt[3] = nyscreen - text_height;


	for (i=0; i<MAXPLAYERS; i++)
	{
		pi = sortedplayers[i];

		if (!playeringame[pi - players]) continue;
		tm = pi->userinfo.team;
		if (tm<0 || tm>=NUM_TEAMS)
		{
			tm = NUM_TEAMS-1;
			color = CR_WHITE;		//Visual indication that something went wrong
		}
		else if (tm>=2)
		{
			color = (pi==player) ?									CR_BRICK :
					(demoplayback && pi==&players[consoleplayer]) ?	CR_BRICK :
																	CR_GRAY;
		}
		else
		{
			color = (pi==player) ?									CR_BRICK :
					(demoplayback && pi==&players[consoleplayer]) ?	CR_BRICK :
																	TeamInfo[tm].color;
		}
		y = yt[tm];
		if (y>=maxyt[tm]) continue;

		if (pi->isbot)	sprintf(str,"%d/-",pi->fragcount);
		else			sprintf(str,"%d/%d",pi->fragcount,pi->ping);

		if (pi->playerstate == PST_WATCHING)
		{
			screen->DrawTextCleanLuc(color,xt[tm],y,pi->userinfo.netname);
			screen->DrawTextCleanLuc(color,xt[tm]+namewidth[tm & 1],y,str);
		}
		else
		{
			screen->DrawTextClean(color,xt[tm],y,pi->userinfo.netname);
			screen->DrawTextClean(color,xt[tm]+namewidth[tm & 1],y,str);
		}

		yt[tm] +=  text_height;
	}
	screen->SetFont(SmallFont);
}


#endif		//0

#endif		//__CTF__

//Lyfe: Coop Scoreboard
static void Draw_CoopScores(player_t *player)
{
	int				i, j, y, maxwidth, nxscreen;
	player_t		*sortedplayers[MAXPLAYERS];
	char			str[256];
	int				text_height = (screen->Font->GetHeight()+1) * CleanYfac;
	
	sortedplayers[MAXPLAYERS-1] = player;
	for (i = 0, j = 0; j < MAXPLAYERS - 1; i++, j++)
	{
		if (&players[i] == player)
			i++;
		sortedplayers[j] = &players[i];
	}

	// sort by the coop kills count, rather than frags or havenot.
	qsort (sortedplayers, MAXPLAYERS, sizeof(player_t *), compare_coopplayers);
	
	nxscreen = SCREENWIDTH;
	
	maxwidth = 0;
	for (i = 0; i < MAXPLAYERS; i++)
	{
		if (playeringame[i])
		{
			int width = screen->StringWidth (players[i].userinfo.netname);
			if (width > maxwidth) maxwidth = width;
		}
	}
	
	y = 2 * CleanYfac;
	y = 0;
	//DPrintf("y = %d -> CleanYfac = %d\n", y, CleanYfac);
	sprintf(str, "Total Monsters: %d\n", level.total_monsters);
	screen->DrawTextClean(CR_GRAY, nxscreen-screen->StringWidth(str) * CleanXfac, 0, str);
	sprintf(str, "Remaining: %d\n", level.total_monsters - level.killed_monsters);
	screen->DrawTextClean(CR_GRAY, nxscreen-screen->StringWidth(str) * CleanXfac, 1*text_height, str);
	sprintf(str, "Map: %s\n", level.mapname);
	screen->DrawTextClean(CR_GRAY, nxscreen-screen->StringWidth(str) * CleanXfac, 2*text_height, str);

	if (timeleft>0)
		screen->DrawTextClean(CR_GRAY, (nxscreen - max_timer_width())/2,y,timer_string());

	//DPrintf("text_height = %d\n", text_height);
	y += text_height + CleanYfac;

	y = (ST_Y >> 1) - (MAXPLAYERS * 6);
	if (y < 48)
		y = 48;

	//DPrintf("y = %d -> y-11*CleanYfac = %d -> (y-11)*CleanYfac = %d\n", y, y - 11 * CleanYfac, (y-11) * CleanYfac);
	// column positions (horizontal)
	// edited to base it on a percent-based scoreboard.
	int positionBoxS        = ( 2.5/100.0) * nxscreen; // color box by username goes from BoxS to BoxE (start to end)
	int positionBoxE        = ( 5.0/100.0) * nxscreen;
	int positionName        = ( 7.5/100.0) * nxscreen;
	int positionHealth      = (42.5/100.0) * nxscreen;
	int positionKills       = (60.0/100.0) * nxscreen;
	int positionKillsPDeath = (72.0/100.0) * nxscreen;
	//int positionItems       = 260 * CleanXfac;
	int positionPing        = (80.5/100.0) * nxscreen;
	int positionTime        = (90.5/100.0) * nxscreen;
	// column headers.
	screen->DrawTextClean (CR_GREY, positionName,        y-11 * CleanYfac, "NAME");
	screen->DrawTextClean (CR_GREY, positionHealth,      y-11 * CleanYfac, "HEALTH");
	screen->DrawTextClean (CR_GREY, positionKills,       y-11 * CleanYfac, "KILLS");
	screen->DrawTextClean (CR_GREY, positionKillsPDeath, y-11 * CleanYfac, "KpD");
	//screen->DrawTextClean (CR_GREY, positionItems,       y-11 * CleanYfac, "Items");
	screen->DrawTextClean (CR_GREY, positionPing,        y-11 * CleanYfac, "Ping");
	screen->DrawTextClean (CR_GREY, positionTime,        y-11 * CleanYfac, "Time");

	int rightadjust = 0;
	for (i = 0; i < MAXPLAYERS && y < ST_Y - 12 * CleanYfac; i++)
	{
		int color = sortedplayers[i]->userinfo.color;

		if (playeringame[sortedplayers[i] - players])
		{
			color = ColorMatcher.Pick (RPART(color), GPART(color), BPART(color));
			int textcolor = (sortedplayers[i] == player) ? CR_GREEN : CR_BRICK;
			bool translucent = (sortedplayers[i]->playerstate == PST_WATCHING) ? true : false;

			// [NightFang] - color box (modified)
			// left , top , right, bottom
			screen->Clear (positionBoxS, y, positionBoxE, y + text_height, color);
			
			// username
			strncpy (str, sortedplayers[i]->userinfo.netname, 16);
			str[17] = '\0';
			if (sortedplayers[i] != player)
				color = (demoplayback && sortedplayers[i] == &players[consoleplayer]) ? CR_GOLD : CR_GREY;
			else
				color = CR_GREEN;
			scoreboardDrawFunction(translucent, color, positionName, y, str);
			
			// health
			sprintf (str, "%3d/%3d", sortedplayers[i]->health, sortedplayers[i]->armorpoints[0]+sortedplayers[i]->armorpoints[1]+sortedplayers[i]->armorpoints[2]);
			rightadjust = positionHealth + (screen->StringWidth("HEALTH") - screen->StringWidth(str)) * CleanXfac;
			scoreboardDrawFunction(translucent, color, rightadjust, y, str);
			
			// number kills
			sprintf (str, "%d", sortedplayers[i]->killcount);
			rightadjust = positionKills + (screen->StringWidth("KILLS") - screen->StringWidth(str)) * CleanXfac;
			scoreboardDrawFunction(translucent, color, rightadjust, y, str);
			
			// kills per death.
			// if killcount == 0, then KpD = 0 (of course).
			// if deathcount == 0, then KpD = "Undef"
			if(sortedplayers[i]->deathcount == 0)
				sprintf(str, "-");
			else
				sprintf(str, "%.1f", (double)sortedplayers[i]->killcount / (double)sortedplayers[i]->deathcount);
			rightadjust = positionKillsPDeath + (screen->StringWidth("KPD") - screen->StringWidth(str)) * CleanXfac;
			scoreboardDrawFunction(translucent, color, rightadjust, y, str);

			// number items
			//sprintf (str, "%d", sortedplayers[i]->itemcount);
			//scoreboardDrawFunction(translucent, color, positionItems, y, str);

			// [NightFang] - ping
			if (sortedplayers[i]->isbot)
				strcpy(str,"-");
			else
				sprintf (str, "%d", sortedplayers[i]->ping);
			rightadjust = positionPing + (screen->StringWidth("Ping") - screen->StringWidth(str)) * CleanXfac;
			scoreboardDrawFunction(translucent, color, rightadjust, y, str);

			// [NightFang] - player time (NEW - not for bots)
			sprintf (str, "%d", sortedplayers[i]->minutesingame);
			rightadjust = positionTime + (screen->StringWidth("Time") - screen->StringWidth(str)) * CleanXfac;
			scoreboardDrawFunction(translucent, color, rightadjust, y, str);

			y += text_height + CleanYfac;
		}
	}
}

static void Draw_DMScores (player_t *player)
{
	int				i, j, y, maxwidth,nxscreen ;
	player_t		*sortedplayers[MAXPLAYERS];
	char			str[256];

	if (player->camera->player)
		player = player->camera->player;

	sortedplayers[MAXPLAYERS-1] = player;
	for (i = 0, j = 0; j < MAXPLAYERS - 1; i++, j++)
	{
		if (&players[i] == player)
			i++;
		sortedplayers[j] = &players[i];
	}

	qsort (sortedplayers, MAXPLAYERS, sizeof(player_t *), compare_players);

	screen->SetFont(ConFont);

	nxscreen = SCREENWIDTH;

	maxwidth = 0;
	for (i = 0; i < MAXPLAYERS; i++)
	{
		if (playeringame[i])
		{
			int width = screen->StringWidth (players[i].userinfo.netname);
			if (width > maxwidth) maxwidth = width;
		}
	}

	y = 2 * CleanYfac;
	if (consoleplayer==displayplayer)
	{
		sprintf(str, "Exp: %d", players[consoleplayer].experience);
		screen->DrawTextClean(CR_GRAY, nxscreen-screen->StringWidth(str) * CleanXfac, y, str);
	}

//-------------------------------------------------
	int	text_height = (screen->Font->GetHeight()+1) * CleanYfac;
	y += text_height + CleanYfac;

	if (timeleft>0)
		screen->DrawTextClean(CR_GRAY,nxscreen - max_timer_width(),y,timer_string());

	int lim1, lim2;

	strcpy(str,"Deathmatch");
	screen->DrawTextClean (CR_GOLD, (nxscreen - screen->StringWidth(str)*CleanXfac)/2,y,str);
	y += text_height+CleanYfac;

	lim1 = (int) fraglimit;
	lim2 = (int) timelimit;
	if (lim1>0 && lim2>0)	sprintf(str,"Frag Limit: %d    Time Limit: %d",lim1, lim2);
	else if (lim1>0)		sprintf(str,"Frag Limit: %d",lim1);
	else if (lim2>0)		sprintf(str,"Time Limit: %d",lim2);
	else					str[0] = '\0';
	if (str[0])
	{
		screen->DrawTextClean(CR_WHITE, (nxscreen - screen->StringWidth(str)*CleanXfac)/2,y,str);
		y += text_height+CleanYfac;
	}

	y += 3*text_height/2;
	screen->DrawTextClean (CR_GREY, 286 * CleanXfac, y-11 * CleanYfac, "Time");
	screen->DrawTextClean (CR_GREY, 248 * CleanXfac, y-11 * CleanYfac, "Ping");
	screen->DrawTextClean (CR_GREY, 198 * CleanXfac, y-11 * CleanYfac, "Frags");
	screen->DrawTextClean (CR_GREY,  34 * CleanXfac, y-11 * CleanYfac, "Name");

	int height = screen->Font->GetHeight() * CleanYfac;

	for (i = 0; i < MAXPLAYERS && y < ST_Y - 12 * CleanYfac; i++)
	{
		int color = sortedplayers[i]->userinfo.color;

		if (playeringame[sortedplayers[i] - players])
		{
			color = ColorMatcher.Pick (RPART(color), GPART(color), BPART(color));

			// [NightFang] - ping
			if (sortedplayers[i]->isbot)	strcpy(str,"-");
			else							sprintf (str, "%d", sortedplayers[i]->ping);
			if (sortedplayers[i]->playerstate == PST_WATCHING )
			{
				screen->DrawTextCleanLuc (sortedplayers[i] == player ? CR_GREEN : CR_BRICK,
								248 * CleanXfac, y, str);
			}
			else
			{
				screen->DrawTextClean (sortedplayers[i] == player ? CR_GREEN : CR_BRICK,
								248 * CleanXfac, y, str);
			}

			// [NightFang] - color box (modified)
			screen->Clear (9*CleanXfac, y, 18 * CleanXfac, y + height, color);
			// left , top , right, bottom
			//screen->Clear (9*CleanXfac, y, 8, 8, color);

			// [NightFang] - frags
			sprintf (str, "%d", sortedplayers[i]->fragcount);
			if (sortedplayers[i]->playerstate == PST_WATCHING)
			{
				screen->DrawTextCleanLuc (sortedplayers[i] == player ? CR_GREEN : CR_BRICK,
								211 * CleanXfac, y, str);
			}
			else
			{
				screen->DrawTextClean (sortedplayers[i] == player ? CR_GREEN : CR_BRICK,
								211 * CleanXfac, y, str);
			}

			// [NightFang] - player time (NEW - not for bots)
			sprintf (str, "%d", sortedplayers[i]->minutesingame);
			if (sortedplayers[i]->playerstate == PST_WATCHING)
			{
				screen->DrawTextCleanLuc (sortedplayers[i] == player ? CR_GREEN : CR_BRICK,
									286 * CleanXfac, y, str);
			}
			else
			{
				screen->DrawTextClean (sortedplayers[i] == player ? CR_GREEN : CR_BRICK,
								286 * CleanXfac, y, str);
			}

			strcpy (str, sortedplayers[i]->userinfo.netname);

			if (sortedplayers[i] != player)
			{
				color = (demoplayback && sortedplayers[i] == &players[consoleplayer]) ? CR_GOLD : CR_GREY;
			}
			else
			{
				color = CR_GREEN;
			}

			// [NightFang] - modified
			if (sortedplayers[i]->playerstate == PST_WATCHING)
				screen->DrawTextCleanLuc (color, 34 * CleanXfac, y, str);
			else
				screen->DrawTextClean (color, 34 * CleanXfac, y, str);

			y += height + CleanYfac;
		}
	}
	screen->SetFont(SmallFont);
}

void HU_DrawScores (player_t *player)
{
	if (ctf)
		Draw_CTFScores(player);			//duke
	else if (teamplay)			
		Draw_TeamDMScores(player);		//Kilgore
	else if (!deathmatch)
		Draw_CoopScores(player);		//Lyfe - Dec 30
	else
		Draw_DMScores(player);
	BorderNeedRefresh = screen->GetPageCount ();
}


static int nparticipants;

void update_rank_spread_info(void)
{
	int		i, n, lastscore, lastrank;

	if (!cl_rankspread)	return;

	if (teamplay)
	{
		int				j, membercount[NUM_TEAMS];
		team_info_s		*t, *sortedteams[NUM_TEAMS];

		memset(membercount,0,sizeof(membercount));
		for (i=0;  i<MAXPLAYERS;  i++)
		{
			if (!playeringame[i] || players[i].playerstate == PST_WATCHING) continue;
			j = players[i].userinfo.team;
			if (j<0 || j>=NUM_TEAMS) continue;
			(membercount[j])++;
		}
		n = 0;
		for (i=0;  i<NUM_TEAMS;  i++)
		{
			if (membercount[i])
				sortedteams[n++] = TeamInfo+i;
		}
		nparticipants = n;
		if (n==1)
		{
			sortedteams[0]->rank = 1;
			sortedteams[0]->spread = 0;
		}
		else if (n>1)
		{
			qsort(sortedteams, n, sizeof(team_info_s *), compare_teams);
			lastscore = -30000;
			for (i=0; i<n; i++)
			{
				t = sortedteams[i];
				if (t->score != lastscore)
				{
					lastscore = t->score;
					lastrank = i+1;
				}
				t->rank = lastrank;
				t->spread = (i==0) ?	t->score - sortedteams[1]->score	:
										t->score - sortedteams[0]->score	;
			}
		}
	}
	else if (deathmatch)
	{
		player_t	*p, *sortedplayers[MAXPLAYERS];

		n = 0;
		for (i=0;  i<MAXPLAYERS;  i++)
		{
			if (!playeringame[i] || players[i].playerstate == PST_WATCHING) continue;
			sortedplayers[n++] = players+i;
		}
		nparticipants = n;
		if (n==1)
		{
			sortedplayers[0]->rank = 1;
			sortedplayers[0]->spread = 0;
		}
		else if (n>1)
		{
			qsort(sortedplayers, n, sizeof(player_t *), compare_players);
			lastscore = -30000;
			for (i=0; i<n; i++)
			{
				p = sortedplayers[i];
				if (p->fragcount != lastscore)
				{
					lastscore = p->fragcount;
					lastrank = i+1;
				}
				p->rank = lastrank;
				p->spread = (i==0) ?	p->fragcount - sortedplayers[1]->fragcount	:
										p->fragcount - sortedplayers[0]->fragcount	;
			}
		}
	}
}

int hud_top()
{
	int y;

	if (con_scaletext)
		y = (!viewactive ? -30 : -10) * CleanYfac;
	else
		y = (!viewactive ? -30 : -10);
	y += (SCREENHEIGHT == realviewheight && viewactive) ? (SCREENHEIGHT + ST_Y)/2 : ST_Y;
	return y;
}


void HU_DrawRankSpread(int i)
{
	player_t	*p;
	int			r, s, nx, y;
	char		temp1[64], temp2[64];

	if (!playeringame[i]) return;

	p = players+i;
	if (p->playerstate == PST_WATCHING) return;
	if (teamplay)
	{
		i = p->userinfo.team;
		if (i<0 || i>=NUM_TEAMS) return;
		r = TeamInfo[i].rank;
		s = TeamInfo[i].spread;
	}
	else if (deathmatch)
	{
		r = p->rank;
		s = p->spread;
	}
	else
	{
		return;
	}
	if (r==0)	return;

	sprintf(temp1, " %d / %d   ", r, nparticipants);
	sprintf(temp2, (s<=0) ? "%d" : "+%d", s);

	screen->SetFont(BigFont);

	y = hud_top()-BigFont->GetHeight()-3;
	nx = screen->StringWidth(temp1);
	screen->DrawText( CR_GOLD, 0, y, temp1);
	screen->DrawText( (s>0) ? CR_GREEN : (s<0) ? CR_RED : CR_GRAY, nx, y, temp2);
	screen->SetFont(SmallFont);
}

void HU_DrawWatchedName()
{
	char *s = players[displayplayer].userinfo.netname;
	screen->DrawText(CR_GOLD,
					(SCREENWIDTH - screen->StringWidth(s))/2,
					hud_top()-SmallFont->GetHeight()-3,s);
}