#ifdef		USEGUI

// zservDlg.cpp : implementation file
//
#include "stdafx.h"
#include "zserv.h"
#include "zservDlg.h"
#include "dmflags.h"
#include <mmsystem.h>
#include <process.h>
#include "errors.h"
#include "doomtype.h"
#include "m_argv.h"
#include "m_misc.h"
#include "d_main.h"
#include "i_system.h"
#include "c_console.h"
#include "c_dispatch.h"
#include "z_zone.h"
#include "version.h"
#include "i_video.h"
#include "stats.h"
#include "sv_main.h"
#include "log.h"
extern	u_long		listening_ip_address;

DArgs				Args;

#define MAX_TERMS	16
void				(STACK_ARGS *TermFuncs[MAX_TERMS])(void);
const char			*TermNames[MAX_TERMS];
static int			NumTerms;

EXTERN_CVAR( String, hostname);
static	bool		zserv32_updatewindow;
static	char		*window_title = "ZServ32 " DETAILED_STRVERSION;
static	char		*ini_name = ".\\zserv32.ini";
static CRichEditCtrl	*g_rec;
static	int			last_player = -1;

class CritSec
{
	public:
	CritSec()		{	InitializeCriticalSection(&cs);	}
	~CritSec()		{	DeleteCriticalSection(&cs);		}
	void enter()	{	EnterCriticalSection(&cs);		}
	void exit()		{	LeaveCriticalSection(&cs);		}
	
	private:
	CRITICAL_SECTION	cs;	
};

static CritSec	cs_playerlist;


void GUI_Init(int i)
{
	player_t *p;

	p = players+i;
	p->newplayer			= false;
	p->deleteplayer			= false;
	p->updateplayerfrags	= false;
	p->updateplayerping		= false;
	p->updateplayertime		= false;
	p->updateplayername		= false;
	p->playerlist			= -1;
	if (i>last_player)	last_player = i;
}

static bool		saved_newplayer, saved_deleteplayer, saved_updateplayerfrags,
				saved_updateplayerping, saved_updateplayername, saved_updateplayertime;
static int		saved_playerlist;

void GUI_Save(player_t *p)
{
	cs_playerlist.enter();
	saved_newplayer			= p->newplayer;
	saved_deleteplayer		= p->deleteplayer;
	saved_updateplayerfrags	= p->updateplayerfrags;
	saved_updateplayerping	= p->updateplayerping;
	saved_updateplayertime	= p->updateplayertime;
	saved_updateplayername	= p->updateplayername;
	saved_playerlist		= p->playerlist;
}

void GUI_Restore(player_t *p)
{
	p->newplayer			= saved_newplayer;
	p->deleteplayer			= saved_deleteplayer;
	p->updateplayerfrags	= saved_updateplayerfrags;
	p->updateplayerping		= saved_updateplayerping;
	p->updateplayertime		= saved_updateplayertime;
	p->updateplayername		= saved_updateplayername;
	p->playerlist			= saved_playerlist;
	cs_playerlist.exit();
}

void GUI_DelPlayer(int i)
{
	cs_playerlist.enter();
	players[i].deleteplayer = true;
	cs_playerlist.exit();
}

void GUI_AddPlayer(int i)
{
	cs_playerlist.enter();
	players[i].newplayer = true;
	cs_playerlist.exit();
}

void GUI_UpdateName(int i)
{
	cs_playerlist.enter();
	players[i].updateplayername = true;
	cs_playerlist.exit();
}

void GUI_UpdateFrags(int i)
{
	cs_playerlist.enter();
	players[i].updateplayerfrags = true;
	cs_playerlist.exit();
}

void GUI_UpdateTime(int i)
{
	cs_playerlist.enter();
	players[i].updateplayertime = true;
	cs_playerlist.exit();
}

void GUI_UpdatePing(int i)
{
	cs_playerlist.enter();
	players[i].updateplayerping = true;
	cs_playerlist.exit();
}

void GUI_UpdateWindow(void)
{
	zserv32_updatewindow = true;
}

void GUI_fputs(const char *s)
{
	if (*s=='\0' || !g_rec) return;
	g_rec->SetSel(-1L,-1L);
	g_rec->ReplaceSel(s,FALSE);
	g_rec->SendMessage(EM_SCROLLCARET);
}

/////////////////////////////////////////////////////////////////////////////
//#include "dbgprintf.cpp"

void addterm (void (STACK_ARGS *func) (), const char *name)
{
	if (NumTerms == MAX_TERMS)
		I_FatalError ("Too many exit functions registered.\nIncrease MAX_TERMS in zservDlg.cpp");
	TermFuncs[NumTerms] = func;
	TermNames[NumTerms] = name;
//dbg_printf("addterm(%d,%s)\n",NumTerms,name);
	NumTerms++;
}

void popterm ()
{
	if (NumTerms)
		NumTerms--;
}

static void STACK_ARGS call_terms (void)
{
	while (NumTerms > 0)
	{
		--NumTerms;
//dbg_printf ("call_terms(%d,%s)\n", NumTerms, TermNames[NumTerms]);
		TermFuncs[NumTerms] ();
	}
}

/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
	virtual BOOL OnInitDialog();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BOOL CAboutDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();
	SetDlgItemText(IDC_ABOUT_VERSION, window_title);
	return TRUE;
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CZservDlg dialog

CZservDlg::CZservDlg(CWnd* pParent /*=NULL*/)
	: CTrayDialog(CZservDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CZservDlg)
	m_Command = _T("");
	m_Title = _T("");
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
}

void CZservDlg::DoDataExchange(CDataExchange* pDX)
{
	CTrayDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CZservDlg)
	DDX_Control(pDX, IDC_PLAYERLIST, m_PlayerList);
	DDX_Text(pDX, IDC_COMMAND, m_Command);
	DDV_MaxChars(pDX, m_Command, 256);
	DDX_Text(pDX, IDC_TITLE, m_Title);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CZservDlg, CTrayDialog)
	//{{AFX_MSG_MAP(CZservDlg)
	ON_WM_SYSCOMMAND()
	ON_WM_QUERYDRAGICON()
	ON_WM_TIMER()
	ON_WM_CLOSE()
	ON_WM_PAINT()
	ON_BN_CLICKED(IDC_SEND, OnSend)
	ON_BN_CLICKED(IDC_SETUPBUTTON, OnSetupbutton)
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CZservDlg message handlers
void SV_RunTics(void);
extern UINT TimerPeriod;
static BOOL please_stop = FALSE;
static BOOL thread_on = FALSE;

unsigned int STACK_ARGS DoDoomThread(LPVOID pParam)
{
	thread_on = TRUE;
	while (!please_stop)
		SV_RunTics();
	thread_on = FALSE;
	return 0;
}

static void stop_thread(void)
{
	please_stop = TRUE;
	for (int i=150; i>0; i--)		// Wait up to 30 secs. for thread to finish
	{
		if (!thread_on) break;
		Sleep(200);
	}
}

static void GetWindowRect_Rel(HWND h,RECT *r)
{
	POINT pt;

	GetWindowRect(h,r);
	int nx = r->right - r->left;
	int ny = r->bottom - r->top;
	pt.x = r->left;
	pt.y = r->top;
	ScreenToClient(GetParent(h),&pt);
	r->left = pt.x;
	r->top = pt.y;
	r->right = r->left + nx;
	r->bottom = r->top + ny;
}

static void fix_layout(HWND hdlg, int cx, int cy)
{
	HWND	h;
	RECT	r;
	int		nx_inner, x_but, y1, y2, x_mar, y_mar, nx_but, ny_but, ny_between;

	h = GetDlgItem(hdlg,IDC_TITLE);
	GetWindowRect_Rel(h,&r);
	x_mar = r.left;
	y_mar = r.top;
	nx_inner = cx - 2*x_mar;
	MoveWindow(h,x_mar,y_mar,nx_inner,r.bottom-r.top,TRUE);
	InvalidateRect(h,NULL,TRUE);

	h = GetDlgItem(hdlg,IDC_PLAYERLIST);
	GetWindowRect_Rel(h,&r);
	MoveWindow(h,x_mar,r.top,nx_inner,r.bottom-r.top,TRUE);
	InvalidateRect(h,NULL,TRUE);
	y1 = r.bottom;

	h = GetDlgItem(hdlg,IDC_SETUPBUTTON);
	GetWindowRect_Rel(h,&r);
	nx_but = r.right-r.left;
	ny_but = r.bottom - r.top;
	x_but = x_mar + nx_inner - nx_but;
	ny_between = r.top - y1;
	MoveWindow(h, x_but, r.top, nx_but, ny_but, TRUE);
	InvalidateRect(h,NULL,TRUE);

	y1 = r.top + ny_but + ny_between;
	y2 = cy - y_mar - ny_but;

	h = GetDlgItem(hdlg,IDC_CONSOLE);
	MoveWindow(h,x_mar,y1,nx_inner,y2-y1-ny_between,TRUE);
	InvalidateRect(h,NULL,TRUE);
	h = GetDlgItem(hdlg,IDC_COMMAND);
	MoveWindow(h,x_mar,y2,nx_inner-nx_but-8,ny_but,TRUE);
	InvalidateRect(h,NULL,TRUE);

	h = GetDlgItem(hdlg,IDC_SEND);
	MoveWindow(h,x_but,y2,nx_but,ny_but,TRUE);
	InvalidateRect(h,NULL,TRUE);
}


static void set_initial_size(HWND hdlg)
{
	RECT	r;
	int		x,y,nx,ny;
	char	buf[128];

	GetPrivateProfileString("GUI Options","WinPos","",buf,sizeof(buf)-1,ini_name);
	if (sscanf(buf,"%d %d %d %d",&x,&y,&nx,&ny)==4 && nx>=150 && ny>=50 &&
		x >= -5 && x<GetSystemMetrics(SM_CXSCREEN)-10 &&
		y >= -5 && y<GetSystemMetrics(SM_CYSCREEN)-10 )
	{
		MoveWindow(hdlg,x,y,nx,ny,TRUE);
	}
	GetClientRect(hdlg,&r);
	fix_layout(hdlg, r.right-r.left, r.bottom-r.top);
}

static void	save_window_size(HWND hdlg)
{
	RECT	r;
	char	buf[128];

	if (IsIconic(hdlg) || IsZoomed(hdlg)) return;
	GetWindowRect(hdlg,&r);
	sprintf(buf,"%d %d %d %d",r.left,r.top,r.right-r.left,r.bottom-r.top);
	WritePrivateProfileString("GUI Options","WinPos", buf,ini_name);
}

void STACK_ARGS write_sentinel(void)
{
	FILE *f;

	if ( (f=fopen("zserv.ok","w")) == NULL ) return;
	fputs("ok\n",f);
	fclose(f);
}

BOOL CZservDlg::OnInitDialog()
{
	CDialog::OnInitDialog();
	TIMECAPS			tc;
	CFont				m_Font;

	SetWindowText(window_title);
	
	Args.SetArgs (__argc, __argv);

	// Set the timer to be as accurate as possible
	if (timeGetDevCaps (&tc, sizeof(tc)) != TIMERR_NOERROR)
		TimerPeriod = 1;	// Assume minimum resolution of 1 ms
	else
		TimerPeriod = tc.wPeriodMin;
	timeBeginPeriod (TimerPeriod);

	atexit (call_terms);
	atterm (I_Quit);

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		CString strAboutMenu;
		strAboutMenu.LoadString(IDS_ABOUTBOX);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	TraySetIcon(IDR_MAINFRAME);
	TraySetToolTip("ZServ32 r0x0rz j00");
	TraySetMenu(IDR_MENU1);
	
	// Create the console window
	ConsoleView.Create(	WS_CHILD|WS_VISIBLE|WS_VSCROLL|WS_HSCROLL|
						ES_MULTILINE|ES_AUTOHSCROLL|ES_AUTOVSCROLL|ES_READONLY,
						CRect(26,243, 376, 330),
						this, IDC_CONSOLE);
	ConsoleView.LimitText(256*1024);
	m_Font.CreateStockObject(DEFAULT_GUI_FONT);
	ConsoleView.SetFont(&m_Font, TRUE);
	ConsoleView.ModifyStyleEx(0,WS_EX_CLIENTEDGE,0);

	g_rec = &ConsoleView;

	// Dont update the text just yet
	GUI_UpdateWindow();

	// Create the colums for our player view
	m_PlayerList.InsertColumn(COL_PLAYER, "Name", LVCFMT_LEFT, 160);
	m_PlayerList.InsertColumn(COL_FRAGS,  "Frags", LVCFMT_LEFT, 45);
	m_PlayerList.InsertColumn(COL_PING,   "Ping", LVCFMT_LEFT, 45);
	m_PlayerList.InsertColumn(COL_TIME,   "Time", LVCFMT_LEFT, 45);
	m_PlayerList.InsertColumn(COL_C2,   "Country", LVCFMT_LEFT, 135);

	m_PlayerList.DeleteAllItems();
	
	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	GUI_fputs("======== ZDaemon " STRVERSION " ========\n\n");

	SetTimer(1,1000,0);			//Kilgore: was 1 (ms); frequent refreshes are a waste of CPU time

	set_initial_size(m_hWnd);
	thread_on = please_stop = FALSE;

	try
	{
		I_DetectOS ();
		D_DoomMain ();
	}
	catch (...)
	{
		//::ExitProcess(1);
		exit(1);
	}

	atterm (write_sentinel);

	if (AfxBeginThread(DoDoomThread, this))
		// Wait up to 30 secs. to ensure that the bgd. thread is on
		for (int i=150; i>0; i--)
		{
			if (thread_on) break;
			Sleep(200);
		}

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CZservDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.DoModal();
	}
	else
	{
		CTrayDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CZservDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CTrayDialog::OnPaint();
	}
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CZservDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}

void CZservDlg::OnTimer(UINT nIDEvent) 
{
	int			i, j, v;
	player_t	*p,*q;
	char		str[320];

	if (zserv32_updatewindow)
	{
		// Set the title (our hostname)
		// Don't display the server's IP address since it
		// can lead to wrong impressions (and questions) if
		// the host is multihomed.
		UCVarValue	servname = hostname.GetGenericRep (CVAR_String);
		char *s1 = NET_AdrToString(&net_local_adr);
		char tooltip_title[128];

		if (listening_ip_address != INADDR_ANY)
		{
			m_Title.Format("'%s' on %s", servname.String, s1);
			sprintf(tooltip_title,"ZServ32 %s on %s",STRVERSION,s1);
		}
		else
		{
			char *s2 = strchr(s1,':');
			if (s2) s2++;
			else s2 = s1;
			m_Title.Format("'%s' on port %s", servname.String, s2);
			sprintf(tooltip_title,"ZServ32 %s on port %s",STRVERSION,s2);
		}
		UpdateData(FALSE);

		// It would be better to display the server's banner in the tooltip;
		// this is useful for admins who run multiple servers on the same 
		// machine. However, there is a 63-char. windows limit on tray 
		// tooltips. So if the banner is too long, we simply display the
		// Zserv version and the listening port/address.
		if (m_Title.GetLength()<64)
			strcpy(tooltip_title, (LPCTSTR)m_Title );
		TraySetToolTip(tooltip_title);

		zserv32_updatewindow = false;
	}

	for (i=0; i<=last_player; i++)
	{
		p = players+i;
		if (!p) continue;
		cs_playerlist.enter();
		if (p->newplayer)
		{
			p->playerlist = v = m_PlayerList.InsertItem(m_PlayerList.GetItemCount(), "");
			p->updateplayername		= true;
			p->updateplayerfrags	= true;
			p->updateplayerping		= true;
			p->updateplayertime		= true;
			p->newplayer			= false;
			const char *c2 = (p->isbot) ? NULL : clients[i].country_name;
			if (!c2) c2 = "";
			m_PlayerList.SetItemText(v, COL_C2, c2 );
		}
		v = p->playerlist;
		if (p->deleteplayer)
		{
			if (v>=0)
				for (j=0;  j<=last_player;  j++)
				{
					q = players+j;
					if (!q) continue;
					if (q->playerlist>v)	q->playerlist--;
				}
			m_PlayerList.DeleteItem(v);
			p->newplayer			= false;
			p->deleteplayer			= false;
			p->updateplayerfrags	= false;
			p->updateplayerping		= false;
			p->updateplayertime		= false;
			p->updateplayername		= false;
			p->playerlist			= -1;
		}
		if (p->updateplayername)
		{ 
			m_PlayerList.SetItemText(v, COL_PLAYER, p->userinfo.netname);
			p->updateplayername = false;
		}
		if (p->updateplayerfrags)
		{ 
			sprintf (str, "%d", p->fragcount);
			m_PlayerList.SetItemText(v, COL_FRAGS, str);
			p->updateplayerfrags = false;
		}
		if (p->updateplayerping)
		{
			sprintf(str, (p->isbot) ? "Bot" : "%d", clients[i].ping);
			m_PlayerList.SetItemText(v, COL_PING, str); 
			p->updateplayerping = false;
		}
		if (p->updateplayertime)
		{
			sprintf (str, "%d", p->minutesingame);
			m_PlayerList.SetItemText(v, COL_TIME, str); 
			p->updateplayertime = false;
		}
		cs_playerlist.exit();

	}
	CTrayDialog::OnTimer(nIDEvent);
}

void CZservDlg::OnClose() 
{
	CTrayDialog::OnClose();
	stop_thread();
	save_window_size(m_hWnd);
}

void CZservDlg::OnSend() 
{
	int			i;
	const char	*pcmd;
	char		toconsole[256];

	UpdateData(TRUE);

	// Convert...char by char
	pcmd = (LPCTSTR) m_Command;
	i = m_Command.GetLength();
	if (i==0) return;

	if (i>sizeof(toconsole)-1) i = sizeof(toconsole)-1;
	memcpy(toconsole,pcmd,i);
	toconsole[i] = '\0';

	// This isnt a command, so we'll just send it out as text
	if (toconsole[0] != '/')
	{
		char	str[256+32];

		sprintf(str, "] CONSOLE [  %s\n", toconsole);

		// Now, send it to all clients
		ZDOP.Init();
		ZDOP.WriteByte(sv_adminchat);
		ZDOP.WriteString(str);
		ZDOP.Broadcast();
		GUI_fputs(str);
	}
	else	// regular command then
	{
		// Send command to the console
		AddCommandString(toconsole+1);
	}
	
	// Clear the editbox
	m_Command.Empty();
	UpdateData(FALSE);
}

void CZservDlg::OnSetupbutton() 
{
	CDmflags	dlg;
	
	dlg.DoModal();
}

BOOL CZservDlg::PreTranslateMessage(MSG* pMsg) 
{
	if ( (pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_ESCAPE) ||
		 (pMsg->message == WM_COMMAND && pMsg->wParam == ID_TRAYMENU_QUIT) )
	{
		if ( MessageBox("Are you sure you want to shut down zserv?",window_title,
						MB_YESNO|MB_ICONQUESTION|MB_DEFBUTTON2) == IDYES )
			GUI_Exit();
		return TRUE;
	}
	return CDialog::PreTranslateMessage(pMsg);
}

void CZservDlg::OnSize(UINT nType, int cx, int cy) 
{
	CTrayDialog::OnSize(nType, cx, cy);
	fix_layout(m_hWnd,cx,cy);	
}

#endif

