/* *** i_debugger.cpp ***
** 
** Written by: Chris Laverdure <dashevil@sympatico.ca>
**
** This file belongs to the public domain.
**
*/

#ifdef EMBDEBUG
#include <stdio.h>
#include "i_instfunc.h"
#include "i_debugger.h"

#define BTLOGIC (bttraverse>(BTLIMIT - 1))?bttraverse-BTLIMIT:bttraverse

//
// char __backtrace_depth(void)
//
// Purpose: Scans the backtrace and calculates depth. (char extrem) keeps track
// 	    of the lowest number hit -- this is for __display_backtrace.
//	    Using the number this function returns it knows how far to indent the
//	    backtrace so when more exits are made than calls everything stays
//	    aligned.
//

char __backtrace_depth(void)
{
   char temp = 0;
   char bttraverse;
   char dist = BTLIMIT;
   char extrem = 0;

   bttraverse = (bt == BTLIMIT) ? 0 : bt + 1;

   while (dist > 0)
   {
	if (btcall[BTLOGIC] == CALL_ENTER)
	   temp++;
	else if (btcall[BTLOGIC] == CALL_EXIT)
	{
	   temp--;
	   if (temp < extrem)
	      extrem = temp;
	}
 
	dist--;
	bttraverse++;
   }

   return extrem;
}

//
// void __display_backtrace(FILE *crashfile)
//
// Purpose: Displays the last functions called and returned from in an indented list for
// 	    "easy reading" (tm).
//

void __display_backtrace(FILE *crashfile)
{
   char bttraverse;
   char dist = BTLIMIT;
   char depth, x;

   depth = __backtrace_depth();
   depth = (depth >= 0) ? 0 : depth * -1;
   bttraverse = (bt == BTLIMIT) ? 0 : bt + 1;

   fprintf(crashfile, "Embedded Debugger Backtrace:\n");

   while (dist > 0)
   {
	fprintf(crashfile, "[%12d] ", btindex[BTLOGIC]);
	for (x = 0; x < (depth + 1); x++) // Indent
	   fprintf(crashfile, " ");

	if (btcall[BTLOGIC] == CALL_ENTER)
	{
	   fprintf(crashfile, "--> ");
           fprintf(crashfile, "Called: %p from: %p\n", btfunc[BTLOGIC], btsite[BTLOGIC]);
	   depth++;
	}
	else if (btcall[BTLOGIC] == CALL_EXIT)
	{
	   fprintf(crashfile, "<-- ");
           fprintf(crashfile, "Exited: %p to: %p\n", btfunc[BTLOGIC], btsite[BTLOGIC]);
	   depth--;
	}
	else
	   fprintf(crashfile, "btcall[%i] == %i\n", BTLOGIC, btcall[BTLOGIC]);

        bttraverse++;
        dist--;
   }

}
#endif
