#if __CTF__

#include "i_sound.h"
#include "v_video.h"
#include "ctf.h"

AActor	*ZD_FindThingById(int id);
int		hud_top();

int		ctf_state[NUM_TEAMS]		= { 0, 0, 0, 0 };
AActor	*ctf_flag_mo[NUM_TEAMS]		= { NULL, NULL, NULL, NULL };
int		ctf_carrier_id[NUM_TEAMS]	= { -1, -1, -1, -1 };

void CTF_DrawHud(void)
{
	int			nteams, x, y, i, j, t, y0, dy, dx;
	bool		states_ok;
	patch_t		*pico[NUM_TEAMS];
	char		temp[32];

	nteams = maxteams;
	if ( W_CheckNumForName("NTCF0") >= 0 )
	{
		states_ok = true;
		for (i=0; i<nteams; i++)
		{
			switch (ctf_state[i])
			{
				case CTF_FS_INBASE:
				case CTF_FS_TAKEN:
					sprintf(temp,"NCTF1%c",'A'+i);
					break;
				case CTF_FS_DROPPED:
					sprintf(temp,"NCTF3%c",'A'+i);
					break;
				default:
					states_ok = false;
					strcpy(temp,"NCTF0");
			}
			pico[i] = (patch_t *) W_CacheLumpName(temp,101);
		}
		dy = pico[0]->height;
		dx = pico[0]->width;

		x = screen->GetWidth() - dx;
		y0 = hud_top() - nteams*dy - 3;

		y = y0;
		for (i=0; i<nteams; i++)
		{
			screen->DrawPatch(pico[i],x,y);
			if (ctf_state[i]==CTF_FS_TAKEN)
			{
				j = ctf_carrier_id[i];
				if (j>=0 && j<MAXPLAYERS && playeringame[j])
				{
					t = players[j].userinfo.team;
					if (t>=0 && t<nteams)
					{
						sprintf(temp,"NCTF2%c",'A'+t);
						screen->DrawPatch((patch_t *)W_CacheLumpName(temp,101),x,y);
					}
				}
			}
			if (displayplayer==ctf_carrier_id[i])
				screen->DrawPatch((patch_t *)W_CacheLumpName("NCTF4",101),x,y);
			y += dy;
		}
	}
	else
	{
		static char *lumps[4][4] = {
				{ "ctf10", "ctf11", "ctf12", "ctf13" },
				{ "ctf20", "ctf21", "ctf22", "ctf23" },
				{ "ctf40", "ctf41", "ctf42", "ctf43" },
				{ "ctf50", "ctf51", "ctf52", "ctf53" }
		};

		for (i=0; i<nteams; i++)
			pico[i] = (patch_t *) W_CacheLumpName(lumps[i][ctf_state[i]],101);

		dy = pico[0]->height;
		dx = pico[0]->width;

		x = screen->GetWidth() - dx;
		y0 = hud_top() - nteams*dy - 3;

		states_ok = true;
		y = y0;
		for (i=0; i<nteams; i++)
		{
			screen->DrawPatch(pico[i],x,y);
			if (displayplayer==ctf_carrier_id[i])
				screen->DrawPatch((patch_t *)W_CacheLumpName("ctf30",101),x,y);
			if (!ctf_state[i])
				states_ok = false;
			y += dy;
		}
	}

	// Draw the team scores only
	// if we know the flag state
	if (states_ok)
	{
		x -= dx;
		y = y0 + dy/4;
		screen->SetFont(BigFont);
		for (i=0; i<nteams; i++)
		{
			sprintf(temp,"%d",TeamInfo[i].score);
			screen->DrawText(TeamInfo[i].text_color,x,y,temp);
			y += dy;
		}
		screen->SetFont(SmallFont);
	}

}

void CTF_Update_Flag_Frames(void)
{
	int		i;
	AActor	*mo;

	for (i=maxteams-1; i>=0; i--)
	{
		mo = ctf_flag_mo[i];
		if (mo)
			mo->SetStateNF( 
				(ctf_state[i]==CTF_FS_TAKEN || ctf_state[i]==CTF_FS_DROPPED) ?
					mo->PainState : mo->SeeState );
	}
}

void CTF_Update_Flag_Pos(int team)
{
	player_t	*p;
	AActor		*flag;
	angle_t		ang, new_angle;
	fixed_t		newx,newy,newz;
	fixed_t		flagdistance = 15;

	p = players + ctf_carrier_id[team];
	if (!p->mo)
		return;

	if ( (flag=ctf_flag_mo[team]) == NULL )
		return;

	new_angle = p->mo->angle;
	ang = new_angle + ANGLE_180;
	newx = p->mo->x + flagdistance*finecosine[ang>>ANGLETOFINESHIFT];
	newy = p->mo->y + flagdistance*finesine[ang>>ANGLETOFINESHIFT];
	newz = p->mo->z;

	if (flag->x != newx || flag->y != newy || flag->z!=newz || flag->angle!=new_angle)
	{
		flag->angle = new_angle;
		flag->SetOrigin(newx,newy,newz);
	}
}

#endif
